"""Module for managing settings."""

from __future__ import annotations

import os
import typing as t
import warnings
from abc import ABCMeta, abstractmethod
from contextlib import contextmanager, suppress
from enum import Enum

import structlog

from meltano.core.setting_definition import (
    SettingDefinition,
    SettingKind,
    SettingMissingError,
)
from meltano.core.settings_store import SettingValueStore
from meltano.core.utils import EnvVarMissingBehavior, flatten
from meltano.core.utils import expand_env_vars as do_expand_env_vars

if t.TYPE_CHECKING:
    from meltano.core.project import Project
    from meltano.core.settings_store import SettingsStoreManager

logger = structlog.stdlib.get_logger(__name__)


# sentinel value to use to prevent leaking sensitive data
REDACTED_VALUE = "(redacted)"

# magic string used as feature flag setting for experimental features
EXPERIMENTAL = "experimental"
FEATURE_FLAG_PREFIX = "ff"


class FeatureFlags(str, Enum):
    """Available Meltano Feature Flags."""

    STRICT_ENV_VAR_MODE = "strict_env_var_mode"
    PLUGIN_LOCKS_REQUIRED = "plugin_locks_required"

    def __str__(self) -> str:
        """Return feature name.

        Returns:
            str: Feature name.
        """
        return self.value

    @property
    def setting_name(self) -> str:
        """Return the setting name for this feature flag.

        Returns:
            The setting name for this feature flag.
        """
        return f"{FEATURE_FLAG_PREFIX}.{self.value}"


class FeatureNotAllowedException(Exception):
    """A disallowed code path is run."""

    def __init__(self, feature) -> None:  # noqa: ANN001
        """Instantiate the error.

        Args:
            feature: the feature flag to check
        """
        super().__init__(feature)
        self.feature = feature

    def __str__(self) -> str:
        """Represent the error as a string.

        Returns:
            string representation of the error
        """
        return f"{self.feature} not enabled."


class SettingsService(metaclass=ABCMeta):
    """Abstract base class for managing settings."""

    LOGGING = False
    supports_environments = True

    def __init__(
        self,
        project: Project,
        *,
        show_hidden: bool = True,
        env_override: dict | None = None,
        config_override: dict | None = None,
    ):
        """Create a new settings service instance.

        Args:
            project: Meltano project instance.
            show_hidden: Whether to display secret setting values.
            env_override: Optional override environment values.
            config_override:  Optional override configuration values.
        """
        self.project = project
        self.show_hidden = show_hidden
        self.env_override = env_override or {}
        self.config_override = config_override or {}
        self._setting_defs: list[SettingDefinition] | None = None

    @property
    @abstractmethod
    def label(self) -> str:
        """Return label.

        Returns:
            Label for the settings service.
        """

    @property
    @abstractmethod
    def docs_url(self):  # noqa: ANN201
        """Return docs URL.

        Returns:
            URL for Meltano doc site.
        """

    @property
    @abstractmethod
    def project_settings_service(self) -> SettingsService:
        """Get a project settings service.

        Returns:
            A ProjectSettingsService
        """

    @property
    def env_prefixes(self) -> list[str]:
        """Return prefixes for setting environment variables.

        Returns:
            prefixes for settings environment variables
        """
        return ["meltano"]

    @property
    @abstractmethod
    def db_namespace(self) -> str:
        """Return namespace for setting value records in system database."""

    @property
    @abstractmethod
    def setting_definitions(self) -> list[SettingDefinition]:
        """Return definitions of supported settings."""

    @property
    def inherited_settings_service(self) -> None:
        """Return settings service to inherit configuration from."""
        return None

    @property
    @abstractmethod
    def meltano_yml_config(self) -> dict:
        """Return current configuration in `meltano.yml`."""

    @abstractmethod
    def update_meltano_yml_config(self, config):  # noqa: ANN001, ANN201
        """Update configuration in `meltano.yml`.

        Args:
            config: updated config
        """

    @abstractmethod
    def process_config(self, config: dict) -> dict:
        """Process configuration dictionary to be used downstream.

        Args:
            config: configuration dictionary to process
        """

    @property
    def flat_meltano_yml_config(self):  # noqa: ANN201
        """Flatten meltano config.

        Returns:
            the flattened config

        """
        return flatten(self.meltano_yml_config, "dot")

    @property
    def env(self):  # noqa: ANN201
        """Return the environment as a dict.

        Returns:
            the environment as a dict.
        """
        return {**os.environ, **self.env_override}

    def config_with_metadata(  # noqa: ANN201
        self,
        prefix: str | None = None,
        extras: bool | None = None,
        source: SettingValueStore = SettingValueStore.AUTO,
        source_manager: SettingsStoreManager | None = None,
        **kwargs,  # noqa: ANN003
    ):
        """Return all config values with associated metadata.

        Args:
            prefix: the prefix for setting names
            extras: extra setting definitions to include
            source: the SettingsStore to use
            source_manager: the SettingsStoreManager to use
            kwargs: additional keyword args to pass during SettingsStoreManager
                instantiation

        Returns:
            dict of config with metadata
        """
        if source_manager:
            # TODO: Ensure only a bulk-enabled manager can be used here
            source_manager.bulk = True  # type: ignore[attr-defined]
        else:
            source_manager = source.manager(self, bulk=True, **kwargs)

        config = {}
        for setting_def in self.definitions(extras=extras):
            if prefix and not setting_def.name.startswith(prefix):
                continue

            value, metadata = self.get_with_metadata(
                setting_def.name,
                setting_def=setting_def,
                source=source,
                source_manager=source_manager,
                **kwargs,
            )

            config[setting_def.name[len(prefix) :] if prefix else setting_def.name] = {
                **metadata,
                "value": value,
            }

        return config

    def as_dict(self, *args, process: bool = False, **kwargs) -> dict:  # noqa: ANN002, ANN003
        """Return settings without associated metadata.

        Args:
            *args: args to pass to config_with_metadata
            process: Whether to process the config
            **kwargs: additional kwargs to pass to config_with_metadata

        Returns:
            dict of name-value settings pairs
        """
        config_metadata = self.config_with_metadata(*args, **kwargs)

        if process:
            config = {
                key: metadata["setting"].post_process_value(metadata["value"])
                for key, metadata in config_metadata.items()
            }
            config = self.process_config(config)
        else:
            config = {
                key: metadata["value"] for key, metadata in config_metadata.items()
            }

        return config

    def as_env(self, *args, **kwargs) -> dict[str, str]:  # noqa: ANN002, ANN003
        """Return settings as an dictionary of environment variables.

        Args:
            *args: args to pass to config_with_metadata
            **kwargs: additional kwargs to pass to config_with_metadata

        Returns:
            settings as environment variables
        """
        env = {}
        for config in self.config_with_metadata(*args, **kwargs).values():
            value = config["value"]
            if value is None:
                continue

            setting_def = config["setting"]
            value = setting_def.stringify_value(value)

            for env_var in self.setting_env_vars(setting_def, for_writing=True):
                if env_var.negated:
                    continue

                env[env_var.key] = value

        return env

    def get_with_metadata(  # noqa: ANN201
        self,
        name: str,
        *,
        redacted: bool = False,
        source: SettingValueStore = SettingValueStore.AUTO,
        source_manager: SettingsStoreManager | None = None,
        setting_def: SettingDefinition | None = None,
        expand_env_vars: bool = True,
        redacted_value: str = REDACTED_VALUE,
        **kwargs,  # noqa: ANN003
    ):
        """Get a setting with associated metadata.

        Args:
            name: the name of the setting to get
            redacted: Whether the setting is redacted
            source: the `SettingsStore` to use
            source_manager: the `SettingsStoreManager` to use
            setting_def: get this `SettingDefinition` instead of name
            expand_env_vars: Whether to expand nested environment variables
            redacted_value: the value to use when redacting the setting
            **kwargs: additional keyword args to pass during
                `SettingsStoreManager` instantiation

        Returns:
            a tuple of the setting value and metadata
        """
        with suppress(SettingMissingError):
            setting_def = setting_def or self.find_setting(name)
        if setting_def:
            name = setting_def.name

        self.log(f"Getting setting '{name}'")

        metadata: dict[str, t.Any] = {
            "name": name,
            "source": source,
            "setting": setting_def,
        }

        expandable_env = {**self.project.dotenv_env, **self.env}
        if setting_def and setting_def.is_extra:
            expandable_env.update(
                self.as_env(
                    extras=False,
                    redacted=redacted,
                    source=source,
                    source_manager=source_manager,
                ),
            )

        manager = source_manager or source.manager(self, **kwargs)
        value, get_metadata = manager.get(name, setting_def=setting_def)
        metadata.update(get_metadata)

        # Can't do conventional SettingsService.feature_flag call to check;
        # it would result in circular dependency
        strict_env_var_mode, _ = source.manager(self.project_settings_service).get(
            f"{FEATURE_FLAG_PREFIX}.{FeatureFlags.STRICT_ENV_VAR_MODE}",
            cast_value=True,
        )
        if expand_env_vars and metadata.get("expandable", False):
            metadata["expandable"] = False
            expanded_value = do_expand_env_vars(
                value,
                env=expandable_env,
                if_missing=EnvVarMissingBehavior(int(strict_env_var_mode)),
            )
            # https://github.com/meltano/meltano/issues/7189#issuecomment-1396112167
            if value and not expanded_value:  # The whole string was missing env vars
                expanded_value = None  # type: ignore[assignment]

            if expanded_value != value:
                metadata["expanded"] = True
                metadata["unexpanded_value"] = value
                value = expanded_value

        if setting_def:
            # Expand flattened config values if the root value is the default
            # or inherited empty object.
            if setting_def.kind == SettingKind.OBJECT and (
                metadata["source"]
                in {SettingValueStore.DEFAULT, SettingValueStore.INHERITED}
            ):
                object_value = {}
                object_source = metadata["source"]
                for setting_key in (setting_def.name, *setting_def.aliases):
                    flat_config_metadata = self.config_with_metadata(
                        prefix=f"{setting_key}.",
                        redacted=redacted,
                        source=source,
                        source_manager=source_manager,
                        expand_env_vars=expand_env_vars,
                    )
                    for nested_key, config_metadata in flat_config_metadata.items():
                        if nested_key in object_value:
                            continue

                        object_value[nested_key] = config_metadata["value"]

                        nested_source = config_metadata["source"]
                        if nested_source.overrides(object_source):
                            object_source = nested_source

                if object_value:
                    value = object_value  # type: ignore[assignment]
                    metadata["source"] = object_source

            cast_value = setting_def.cast_value(value)
            if cast_value != value:
                metadata["uncast_value"] = value
                value = cast_value

            # we don't want to leak secure informations
            # so we redact all `passwords`
            if redacted and value and setting_def.is_redacted:
                metadata["redacted"] = True
                value = redacted_value

        self.log(f"Got setting {name!r} with metadata: {metadata}")

        if setting_def is None and metadata["source"] is SettingValueStore.DEFAULT:
            warnings.warn(
                (
                    f"Unknown setting {name!r} - the default value "
                    f"`{value!r}` will be used"
                ),
                RuntimeWarning,
                stacklevel=2,
            )

        return value, metadata

    def get_with_source(self, *args, **kwargs):  # noqa: ANN002, ANN003, ANN201
        """Get a setting value along with its source.

        Args:
            *args: args to pass to get_with_metadata
            **kwargs: kwargs to pass to get_with_metadata

        Returns:
            tuple of setting value and its source
        """
        value, metadata = self.get_with_metadata(*args, **kwargs)
        return value, metadata["source"]

    def get(self, *args, **kwargs):  # noqa: ANN002, ANN003, ANN201
        """Get a setting value.

        Args:
            *args: args to pass to get_with_metadata
            **kwargs: kwargs to pass to get_with_metadata

        Returns:
            the setting value
        """
        value, _ = self.get_with_source(*args, **kwargs)
        return value

    def set_with_metadata(  # noqa: ANN201
        self,
        path: str | list[str],
        value,  # noqa: ANN001
        store=SettingValueStore.AUTO,  # noqa: ANN001
        *,
        redacted_value: str = REDACTED_VALUE,
        **kwargs,  # noqa: ANN003
    ):
        """Set the value and metadata for a setting.

        Args:
            path: the key for the setting
            value: the value to set the setting to
            store: the store to set the value in
            redacted_value: the value to use when redacting the setting
            **kwargs: additional keyword args to pass during
                `SettingsStoreManager` instantiation

        Returns:
            the new value and metadata for the setting
        """
        self.log(f"Setting setting '{path}'")

        if isinstance(path, str):
            path = [path]

        name = ".".join(path)

        try:
            setting_def = self.find_setting(name)
        except SettingMissingError:
            warnings.warn(f"Unknown setting {name!r}", RuntimeWarning, stacklevel=2)
            setting_def = None

        metadata = {"name": name, "path": path, "store": store, "setting": setting_def}

        if value == redacted_value:
            metadata["redacted"] = True
            return None, metadata

        if setting_def:
            cast_value = setting_def.cast_value(value)
            if cast_value != value:
                metadata["uncast_value"] = value
                value = cast_value

        metadata.update(
            store.manager(self, **kwargs).set(
                name,
                path,
                value,
                setting_def=setting_def,
            ),
        )

        self.log(f"Set setting {name!r} with metadata: {metadata}")
        return value, metadata

    def set(self, *args, **kwargs):  # noqa: ANN002, ANN003, ANN201
        """Set the value for a setting.

        Args:
            *args: args to pass to set_with_metadata
            **kwargs: kwargs to pass to set_with_metadata

        Returns:
            the new value for the setting
        """
        value, _ = self.set_with_metadata(*args, **kwargs)
        return value

    def unset(self, path: list[str], store=SettingValueStore.AUTO, **kwargs):  # noqa: ANN001, ANN003, ANN201
        """Unset a setting.

        Args:
            path: the key for the setting
            store: the store to set the value in
            **kwargs: additional keyword args to pass during
                SettingsStoreManager instantiation

        Returns:
            the metadata for the setting
        """
        self.log(f"Unsetting setting '{path}'")

        if isinstance(path, str):
            path = [path]

        name = ".".join(path)

        try:
            setting_def = self.find_setting(name)
        except SettingMissingError:
            setting_def = None

        metadata = {
            "name": name,
            "path": path,
            "store": store,
            "setting": setting_def,
            **store.manager(self, **kwargs).unset(name, path, setting_def=setting_def),
        }

        self.log(f"Unset setting {name!r} with metadata: {metadata}")
        return metadata

    def reset(self, store=SettingValueStore.AUTO, **kwargs):  # noqa: ANN001, ANN003, ANN201
        """Reset a setting.

        Args:
            store: the store to set the value in
            **kwargs: additional keyword args to pass during
                `SettingsStoreManager` instantiation

        Returns:
            the metadata for the setting
        """
        metadata = {"store": store, **store.manager(self, **kwargs).reset()}
        self.log(f"Reset settings with metadata: {metadata}")
        return metadata

    def definitions(self, extras: bool | None = None) -> t.Iterable[SettingDefinition]:
        """Return setting definitions along with extras.

        Args:
            extras: additional settings to return

        Returns:
            list of setting definitions
        """
        if self._setting_defs is None:
            self._setting_defs = [
                setting
                for setting in self.setting_definitions
                if not setting.hidden or self.show_hidden
            ]

        if extras is not None:
            return [
                setting
                for setting in self._setting_defs
                if (extras is True and setting.is_extra)
                or (extras is False and not setting.is_extra)
            ]

        return self._setting_defs

    def find_setting(self, name: str) -> SettingDefinition:
        """Find a setting by name.

        Args:
            name:the name or alias of the setting to return

        Returns:
            the setting definition matching the given name

        Raises:
            SettingMissingError: if the setting is not found

        """
        try:
            return next(
                setting
                for setting in self.definitions()
                if setting.name == name or name in setting.aliases
            )
        except StopIteration as err:
            raise SettingMissingError(name) from err

    # TODO: The `for_writing` parameter is unused, but referenced elsewhere.
    # Callers should be updated to not use it, and then it should be removed.
    def setting_env_vars(  # noqa: ANN201
        self,
        setting_def,  # noqa: ANN001
        *,
        for_writing=False,  # noqa: ANN001, ARG002
    ):
        """Get environment variables for the given setting definition.

        Args:
            setting_def: The setting definition to get env vars for.
            for_writing: Unused parameter.

        Returns:
            Environment variables for given setting
        """
        return setting_def.env_vars(self.env_prefixes)

    def setting_env(self, setting_def):  # noqa: ANN001, ANN201
        """Get a single environment variable for the given setting definition.

        Args:
            setting_def: the setting definition to get env vars for

        Returns:
            environment variable for given setting
        """
        return self.setting_env_vars(setting_def)[0].key

    def log(self, message) -> None:  # noqa: ANN001
        """Log the given message.

        Args:
            message: the message to log
        """
        if self.LOGGING:
            logger.debug(message)

    @contextmanager
    def feature_flag(
        self,
        feature: str,
        *,
        raise_error: bool = True,
    ) -> t.Generator[bool, None, None]:
        """Gate code paths based on feature flags.

        Args:
            feature: the feature flag to check
            raise_error: indicates whether error should be raised

        Yields:
            true if the feature flag is enabled, else false

        Raises:
            FeatureNotAllowedException: if `raise_error` is `True` and feature
                flag is disallowed
        """
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", "Unknown setting", RuntimeWarning)

            # experimental is a top-level setting
            if feature == EXPERIMENTAL:
                allowed = self.get(EXPERIMENTAL) or False
            # other feature flags are nested under feature flag
            else:
                allowed = self.get(f"{FEATURE_FLAG_PREFIX}.{feature}") or False

        try:
            yield t.cast(bool, allowed)
        finally:
            if raise_error and not allowed:
                raise FeatureNotAllowedException(feature)



"""Install plugins into the project, using pip in separate venv by default."""

from __future__ import annotations

import asyncio
import functools
import os
import shlex
import sys
import typing as t
from dataclasses import dataclass
from enum import Enum
from functools import cached_property
from multiprocessing import cpu_count

import structlog

from meltano.core.error import (
    AsyncSubprocessError,
    PluginInstallError,
    PluginInstallWarning,
)
from meltano.core.plugin import PluginType
from meltano.core.plugin.settings_service import PluginSettingsService
from meltano.core.settings_service import FeatureFlags
from meltano.core.utils import (
    EnvironmentVariableNotSetError,
    EnvVarMissingBehavior,
    expand_env_vars,
    noop,
)
from meltano.core.venv_service import (
    UvVenvService,
    VenvService,
    VirtualEnv,
    fingerprint,
)

if t.TYPE_CHECKING:
    from meltano.core.plugin.project_plugin import ProjectPlugin
    from meltano.core.project import Project

logger = structlog.stdlib.get_logger(__name__)


class PluginInstallReason(str, Enum):
    """Plugin install reason enum."""

    ADD = "add"
    AUTO = "auto"
    INSTALL = "install"
    UPGRADE = "upgrade"


class PluginInstallStatus(Enum):
    """The status of the process of installing a plugin."""

    RUNNING = "running"
    SUCCESS = "success"
    SKIPPED = "skipped"
    ERROR = "error"
    WARNING = "warning"


@dataclass(frozen=True)
class PluginInstallState:
    """A message reporting the progress of installing a plugin.

    plugin: Plugin related to this install state.
    reason: Reason for plugin install.
    status: Status of plugin install.
    message: Formatted install state message.
    details: Extra details relating to install (including error details if failed).
    """

    plugin: ProjectPlugin
    reason: PluginInstallReason
    status: PluginInstallStatus
    message: str | None = None
    details: str | None = None

    @cached_property
    def successful(self) -> bool:
        """Plugin install success status.

        Returns:
            `True` if plugin install successful.
        """
        return self.status in {PluginInstallStatus.SUCCESS, PluginInstallStatus.SKIPPED}

    @cached_property
    def skipped(self) -> bool:
        """Plugin install skipped status.

        Returns:
            `True` if the installation was skipped / not needed.
        """
        return self.status == PluginInstallStatus.SKIPPED

    @cached_property
    def verb(self) -> str:
        """Verb form of status.

        Returns:
            Verb form of status.
        """
        if self.status is PluginInstallStatus.RUNNING:
            return (
                "Updating"
                if self.reason is PluginInstallReason.UPGRADE
                else "Installing"
            )
        if self.status is PluginInstallStatus.SUCCESS:
            return (
                "Updated" if self.reason is PluginInstallReason.UPGRADE else "Installed"
            )
        if self.status is PluginInstallStatus.SKIPPED:
            return "Skipped installing"

        return "Errored"


def with_semaphore(func):  # noqa: ANN001, ANN201
    """Gate access to the method using its class's semaphore.

    Args:
        func: Function to wrap.

    Returns:
        Wrapped function.
    """

    @functools.wraps(func)
    async def wrapper(self, *args, **kwargs):  # noqa: ANN001, ANN002, ANN003, ANN202
        async with self.semaphore:
            return await func(self, *args, **kwargs)

    return wrapper


class PluginInstallService:
    """Plugin install service."""

    def __init__(
        self,
        project: Project,
        status_cb: t.Callable[[PluginInstallState], t.Any] = noop,
        *,
        parallelism: int | None = None,
        clean: bool = False,
        force: bool = False,
    ):
        """Initialize new PluginInstallService instance.

        Args:
            project: Meltano Project.
            status_cb: Status call-back function.
            parallelism: Number of parallel installation processes to use.
            clean: Clean install flag.
            force: Whether to ignore the Python version required by plugins.
        """
        self.project = project
        self.status_cb = status_cb
        self._parallelism = parallelism
        self.clean = clean
        self.force = force

    @cached_property
    def parallelism(self) -> int:
        """Return the number of parallel installation processes to use.

        Returns:
            The number of parallel installation processes to use.
        """
        if self._parallelism is None:
            return cpu_count()
        if self._parallelism < 1:
            return sys.maxsize
        return self._parallelism

    @cached_property
    def semaphore(self) -> asyncio.Semaphore:
        """An asyncio semaphore with a counter starting at `self.parallelism`.

        Returns:
            An asyncio semaphore with a counter starting at `self.parallelism`.
        """
        return asyncio.Semaphore(self.parallelism)

    @staticmethod
    def remove_duplicates(
        plugins: t.Iterable[ProjectPlugin],
        reason: PluginInstallReason,
    ) -> tuple[list[PluginInstallState], list[ProjectPlugin]]:
        """Deduplicate list of plugins, keeping the last occurrences.

        Trying to install multiple plugins into the same venv via `asyncio.run`
        will fail due to a race condition between the duplicate installs. This
        is particularly problematic if `clean` is set as one async `clean`
        operation causes the other install to fail.

        Args:
            plugins: An iterable containing plugins to dedupe.
            reason: Plugins install reason.

        Returns:
            A tuple containing a list of PluginInstallState instance (for
            skipped plugins) and a deduplicated list of plugins to install.
        """
        seen_venvs = set()
        deduped_plugins: list[ProjectPlugin] = []
        states: list[PluginInstallState] = []
        for plugin in plugins:
            if (plugin.type, plugin.venv_name) not in seen_venvs:
                deduped_plugins.append(plugin)
                seen_venvs.add((plugin.type, plugin.venv_name))
            else:
                states.append(
                    PluginInstallState(
                        plugin=plugin,
                        reason=reason,
                        status=PluginInstallStatus.SKIPPED,
                        message=(
                            f"Plugin {plugin.name!r} does not require "
                            "installation: reusing parent virtualenv"
                        ),
                    ),
                )
        return states, deduped_plugins

    async def install_all_plugins(
        self,
        reason: PluginInstallReason = PluginInstallReason.INSTALL,
    ) -> list[PluginInstallState]:
        """Install all the plugins for the project.

        Blocks until all plugins are installed.

        Args:
            reason: Plugin install reason.

        Returns:
            Install state of installed plugins.
        """
        return await self.install_plugins(self.project.plugins.plugins(), reason=reason)

    async def install_plugins(
        self,
        plugins: t.Iterable[ProjectPlugin],
        reason: PluginInstallReason = PluginInstallReason.INSTALL,
    ) -> list[PluginInstallState]:
        """Install all the provided plugins.

        Args:
            plugins: ProjectPlugin instances to install.
            reason: Plugin install reason.

        Returns:
            Install state of installed plugins.
        """
        states, new_plugins = self.remove_duplicates(plugins=plugins, reason=reason)
        for state in states:
            self.status_cb(state)

        installing = [
            self.install_plugin_async(plugin, reason) for plugin in new_plugins
        ]

        states.extend(await asyncio.gather(*installing))
        return states

    def install_plugin(
        self,
        plugin: ProjectPlugin,
        reason: PluginInstallReason = PluginInstallReason.INSTALL,
    ) -> PluginInstallState:
        """Install a plugin.

        Blocks until the plugin is installed.

        Args:
            plugin: ProjectPlugin to install.
            reason: Install reason.

        Returns:
            PluginInstallState state instance.
        """
        return asyncio.run(
            self.install_plugin_async(
                plugin,
                reason=reason,
            ),
        )

    @with_semaphore
    async def install_plugin_async(
        self,
        plugin: ProjectPlugin,
        reason: PluginInstallReason = PluginInstallReason.INSTALL,
    ) -> PluginInstallState:
        """Install a plugin asynchronously.

        Args:
            plugin: ProjectPlugin to install.
            reason: Install reason.

        Returns:
            PluginInstallState state instance.
        """
        env = self.plugin_installation_env(plugin)

        if (
            (
                reason == PluginInstallReason.AUTO
                and not self._requires_install(plugin, env=env)
            )
            or not plugin.is_installable()
            or self._is_mapping(plugin)
        ):
            state = PluginInstallState(
                plugin=plugin,
                reason=reason,
                status=PluginInstallStatus.SKIPPED,
                message=f"Plugin '{plugin.name}' does not require installation",
            )
            self.status_cb(state)
            return state

        self.status_cb(
            PluginInstallState(
                plugin=plugin,
                reason=reason,
                status=PluginInstallStatus.RUNNING,
            ),
        )

        try:
            async with plugin.trigger_hooks("install", self, plugin, reason):
                installer: PluginInstaller = getattr(
                    plugin,
                    "installer",
                    install_pip_plugin,
                )
                await installer(
                    project=self.project,
                    plugin=plugin,
                    reason=reason,
                    clean=self.clean,
                    force=self.force,
                    env=env,
                )
                state = PluginInstallState(
                    plugin=plugin,
                    reason=reason,
                    status=PluginInstallStatus.SUCCESS,
                )
                self.status_cb(state)
                return state

        except PluginInstallError as err:
            state = PluginInstallState(
                plugin=plugin,
                reason=reason,
                status=PluginInstallStatus.ERROR,
                message=str(err),
            )
            self.status_cb(state)
            return state

        except PluginInstallWarning as warn:
            state = PluginInstallState(
                plugin=plugin,
                reason=reason,
                status=PluginInstallStatus.WARNING,
                message=str(warn),
            )
            self.status_cb(state)
            return state

        except AsyncSubprocessError as err:
            state = PluginInstallState(
                plugin=plugin,
                reason=reason,
                status=PluginInstallStatus.ERROR,
                message=(
                    f"{plugin.type.descriptor.capitalize()} '{plugin.name}' "
                    f"could not be installed: {err}"
                ),
                details=await err.stderr,
            )
            self.status_cb(state)
            return state

    def _requires_install(
        self,
        plugin: ProjectPlugin,
        *,
        env: t.Mapping[str, str] | None = None,
    ) -> bool:
        try:
            pip_install_args = get_pip_install_args(
                self.project,
                plugin,
                env,
                if_missing=EnvVarMissingBehavior.raise_exception,
            )
        except EnvironmentVariableNotSetError as e:
            logger.warning(
                (
                    "Environment variable '%s' not set for '%s' `pip_url`, will not"
                    " attempt install"
                ),
                e.env_var,
                plugin.name,
            )
            return False

        venv = VirtualEnv(self.project.plugin_dir(plugin, "venv", make_dirs=False))
        return fingerprint(pip_install_args) != venv.read_fingerprint()

    @staticmethod
    def _is_mapping(plugin: ProjectPlugin) -> bool:
        """Check if a plugin is a mapping, as mappings are not installed.

        Mappings are `PluginType.MAPPERS` with extra attribute of `_mapping`
        which will indicate that this instance of the plugin is actually a
        mapping - and should not be installed.

        Args:
            plugin: ProjectPlugin to evaluate.

        Returns:
            A boolean determining if the given plugin is a mapping (of type
            `PluginType.MAPPERS`).
        """
        return plugin.type == PluginType.MAPPERS and bool(
            plugin.extra_config.get("_mapping")
        )

    def plugin_installation_env(self, plugin: ProjectPlugin) -> dict[str, str]:
        """Environment variables to use during plugin installation.

        Args:
            plugin: The plugin being installed.

        Returns:
            A dictionary of environment variables from the process'
            environment, `meltano.yml`, the plugin `env` config, et cetera, in
            accordance with the normal Meltano env precedence hierarchy. See
            https://docs.meltano.com/guide/configuration#specifying-environment-variables.
            A special env var (with lowest precedence) `$MELTANO__PYTHON_VERSION`
            is included, and has the value
            `<major Python version>.<minor Python version>`.
        """
        plugin_settings_service = PluginSettingsService(self.project, plugin)
        with self.project.settings.feature_flag(
            FeatureFlags.STRICT_ENV_VAR_MODE,
            raise_error=False,
        ) as strict_env_var_mode:
            expanded_project_env = expand_env_vars(
                self.project.settings.env,
                os.environ,
                if_missing=EnvVarMissingBehavior(strict_env_var_mode),
            )
            return {
                "MELTANO__PYTHON_VERSION": (
                    f"{sys.version_info.major}.{sys.version_info.minor}"
                ),
                **expanded_project_env,
                **expand_env_vars(
                    plugin_settings_service.project.dotenv_env,
                    os.environ,
                    if_missing=EnvVarMissingBehavior(strict_env_var_mode),
                ),
                **plugin_settings_service.as_env(),
                **plugin_settings_service.plugin.info_env,
                **expand_env_vars(
                    plugin_settings_service.plugin.env,
                    expanded_project_env,
                    if_missing=EnvVarMissingBehavior(strict_env_var_mode),
                ),
            }


class PluginInstaller(t.Protocol):
    """Prototype function for plugin installation.

    All plugin installation functions must support at least the specified
    parameters, and also accept additional unused keyword arguments.
    """

    async def __call__(
        self,
        *,
        project: Project,
        plugin: ProjectPlugin,
        **kwargs,  # noqa: ANN003
    ) -> None:
        """Install the plugin.

        Args:
            project: Meltano Project.
            plugin: `ProjectPlugin` to install.
            kwargs: Additional arguments for the installation of the plugin.
        """


def get_pip_install_args(
    project: Project,
    plugin: ProjectPlugin,
    env: t.Mapping[str, str] | None = None,
    if_missing: EnvVarMissingBehavior | None = None,
) -> list[str]:
    """Get the pip install arguments for the given plugin.

    Args:
        project: Meltano Project.
        plugin: `ProjectPlugin` to get pip install arguments for.
        env: Optional environment variables to use when expanding the pip install args.
        if_missing: The behaviour flow to follow when a environment variable is not
            present when expanding the pip URL

    Returns:
        The list of pip install arguments for the given plugin.
    """
    with project.settings.feature_flag(
        FeatureFlags.STRICT_ENV_VAR_MODE,
        raise_error=False,
    ) as strict_env_var_mode:
        return shlex.split(
            expand_env_vars(
                plugin.pip_url or "",
                env or {},
                if_missing=if_missing or EnvVarMissingBehavior(strict_env_var_mode),
            )
            or "",
        )


async def install_pip_plugin(
    *,
    project: Project,
    plugin: ProjectPlugin,
    clean: bool = False,
    force: bool = False,
    env: t.Mapping[str, str] | None = None,
    **kwargs,  # noqa: ANN003, ARG001
) -> None:
    """Install the plugin with pip.

    Args:
        project: Meltano Project.
        plugin: `ProjectPlugin` to install.
        clean: Flag to clean install.
        force: Whether to ignore the Python version required by plugins.
        env: Environment variables to use when expanding the pip install args.
        kwargs: Unused additional arguments for the installation of the plugin.

    Raises:
        ValueError: If the venv backend is not supported.
    """
    pip_install_args = get_pip_install_args(project, plugin, env=env)
    backend = project.settings.get("venv.backend")

    if backend == "virtualenv":
        service = VenvService(
            project=project,
            python=plugin.python,
            namespace=plugin.type,
            name=plugin.venv_name,
        )
    elif backend == "uv":  # pragma: no cover
        service = UvVenvService(
            project=project,
            python=plugin.python,
            namespace=plugin.type,
            name=plugin.venv_name,
        )
    else:  # pragma: no cover
        msg = f"Unsupported venv backend: {backend}"
        raise ValueError(msg)

    await service.install(
        pip_install_args=("--ignore-requires-python", *pip_install_args)
        if force
        else pip_install_args,
        clean=clean,
        env={
            **os.environ,
            **project.dotenv_env,
        },
    )



"""`BlockSet` metaclass and related components."""

from __future__ import annotations

import typing as t
from abc import ABCMeta, abstractmethod

if t.TYPE_CHECKING:
    from meltano.core.block.ioblock import IOBlock


class BlockSetValidationError(Exception):
    """Base exception when a block in a BlockSet violates the sets requirements."""

    def __init__(self, error: str, message: str = "block violates set requirements"):
        """Initialize exception for when plugin violates the BlockSet requirements.

        Args:
            error: The error.
            message: The message.
        """
        super().__init__(f"{message}: {error}")


class BlockSet(metaclass=ABCMeta):
    """Currently the only complex block set is our `ExtractLoadBlocks` type.

    Theoretically, this is the bare minimum that we need to run and terminate
    (i.e. early abort) a block set. So anything implementing a `run`,
    `terminate`, and `validate_set` method currently satisfies the `BlockSet`
    interface.
    """

    blocks: tuple[IOBlock, ...]

    @abstractmethod
    async def run(self) -> None:
        """Do whatever a BlockSet is designed to do."""
        raise NotImplementedError

    @abstractmethod
    async def terminate(self, *, graceful: bool = True) -> t.NoReturn:
        """Terminate a currently executing BlockSet.

        Args:
            graceful: Whether the BlockSet should try to gracefully quit.
        """
        raise NotImplementedError

    @abstractmethod
    def validate_set(self) -> None:
        """Validate a block set to ensure its valid and runnable.

        Raises:
            BlockSetValidationError: on validation failure
        """
        raise NotImplementedError



from __future__ import annotations  # noqa: D104


class RunnerError(Exception):  # noqa: D101
    def __init__(self, message, exitcodes=None) -> None:  # noqa: ANN001, D107
        super().__init__(message)
        self.exitcodes = {} if exitcodes is None else exitcodes


class Runner:  # noqa: D101
    def run(self) -> None:  # noqa: D102
        pass



"""This module contains the SingerTap class as well as a supporting methods."""

from __future__ import annotations

import asyncio
import json
import logging
import shutil
import sys
import typing as t
from contextlib import suppress
from functools import lru_cache, reduce
from hashlib import sha1
from io import StringIO
from itertools import takewhile

import structlog
from jsonschema import Draft4Validator

from meltano.core.behavior.hookable import hook
from meltano.core.plugin.error import PluginExecutionError, PluginLacksCapabilityError
from meltano.core.setting_definition import SettingDefinition, SettingKind
from meltano.core.state_service import SINGER_STATE_KEY, StateService
from meltano.core.utils import file_has_data, flatten

from . import PluginType, SingerPlugin
from .catalog import (
    CatalogDict,
    MetadataExecutor,
    MetadataRule,
    SchemaExecutor,
    SchemaRule,
    property_breadcrumb,
    select_filter_metadata_rules,
    select_metadata_rules,
)

if t.TYPE_CHECKING:
    from asyncio.streams import StreamReader
    from pathlib import Path

    from meltano.core.plugin_invoker import PluginInvoker

logger = structlog.getLogger(__name__)


async def _stream_redirect(
    stream: asyncio.StreamReader | None,
    *file_like_objs,  # noqa: ANN002
    write_str=False,  # noqa: ANN001
) -> None:
    """Redirect stream to a file like obj.

    Args:
        stream: the stream to redirect
        file_like_objs: the objects to redirect the stream to
        write_str: if True, stream is written as str
    """
    encoding = sys.getdefaultencoding()
    while stream and not stream.at_eof():
        data = await stream.readline()
        for file_like_obj in file_like_objs:
            file_like_obj.write(data.decode(encoding) if write_str else data)


def _debug_logging_handler(
    name: str,
    plugin_invoker: PluginInvoker,
    stderr: StreamReader,
    *other_dsts,  # noqa: ANN002
) -> asyncio.Task:
    """Route debug log lines.

    Routes to stderr, or an `OutputLogger` if one is present in our invocation
    context.

    Args:
        name: name of the plugin
        plugin_invoker: the PluginInvoker to route log lines for
        stderr: stderr StreamReader to route to
        other_dsts: other destinations that the stream should be routed too
            along with logging output

    Returns:
        asyncio.Task which performs the routing of log lines
    """
    if not plugin_invoker.context or not plugin_invoker.context.base_output_logger:
        return asyncio.ensure_future(
            _stream_redirect(
                stderr,
                sys.stderr,
                *other_dsts,
                write_str=True,
            ),
        )

    out = plugin_invoker.context.base_output_logger.out(
        name,
        logger.bind(type="discovery", stdio="stderr"),
    )
    with out.line_writer() as outerr:
        return asyncio.ensure_future(
            _stream_redirect(
                stderr,
                outerr,
                *other_dsts,
                write_str=True,
            ),
        )


def config_metadata_rules(config):  # noqa: ANN001, ANN201
    """Get metadata rules from config.

    Args:
        config: configuration dict

    Returns:
        a list of MetadataRule
    """
    flat_config = flatten(config, "dot")

    rules = []
    for key, value in flat_config.items():
        # <tap_stream_id>.<key>
        # <tap_stream_id>.<prop>.<key>
        # <tap_stream_id>.<prop>.<subprop>.<key>
        # <tap_stream_id>.properties.<prop>.<key>
        # <tap_stream_id>.properties.<prop>.properties.<subprop>.<key>
        tap_stream_id, *props, rule_key = key.split(".")

        rules.append(
            MetadataRule(
                tap_stream_id=tap_stream_id,
                breadcrumb=property_breadcrumb(props),
                key=rule_key,
                value=value,
            ),
        )

    return rules


def config_schema_rules(config):  # noqa: ANN001, ANN201
    """Get schema rules from config.

    Args:
        config: configuration dict

    Returns:
        a list of SchemaRule
    """
    return [
        SchemaRule(
            tap_stream_id=tap_stream_id,
            breadcrumb=["properties", prop],
            payload=payload,
        )
        for tap_stream_id, stream_config in config.items()
        for prop, payload in stream_config.items()
    ]


class SingerTap(SingerPlugin):
    """A Plugin for Singer Taps."""

    __plugin_type__ = PluginType.EXTRACTORS

    EXTRA_SETTINGS: t.ClassVar[list[SettingDefinition]] = [
        SettingDefinition(name="_catalog"),
        SettingDefinition(name="_state"),
        SettingDefinition(name="_load_schema", value="$MELTANO_EXTRACTOR_NAMESPACE"),
        SettingDefinition(name="_select", kind=SettingKind.ARRAY, value=["*.*"]),
        SettingDefinition(
            name="_metadata",
            aliases=["metadata"],
            kind=SettingKind.OBJECT,
            value={},
            value_processor="nest_object",
        ),
        SettingDefinition(
            name="_schema",
            kind=SettingKind.OBJECT,
            value={},
            value_processor="nest_object",
        ),
        SettingDefinition(name="_select_filter", kind=SettingKind.ARRAY, value=[]),
        SettingDefinition(
            name="_use_cached_catalog",
            kind=SettingKind.BOOLEAN,
            value=True,
        ),
    ]

    def exec_args(self, plugin_invoker):  # noqa: ANN001, ANN201
        """Return the arguments list with the complete runtime paths.

        Args:
            plugin_invoker: the plugin invoker running

        Returns:
            the command line arguments to be passed to the tap
        """
        args = ["--config", plugin_invoker.files["config"]]

        catalog_path = plugin_invoker.files["catalog"]
        if file_has_data(catalog_path):
            if "catalog" in plugin_invoker.capabilities:
                args += ["--catalog", catalog_path]
            elif "properties" in plugin_invoker.capabilities:
                args += ["--properties", catalog_path]
            else:
                logger.warning(
                    "A catalog file was found, but it will be ignored as the "
                    "extractor does not advertise the `catalog` or "
                    "`properties` capability",
                )

        state_path = plugin_invoker.files["state"]
        if file_has_data(state_path):
            if "state" in plugin_invoker.capabilities:
                args += ["--state", state_path]
            else:
                logger.warning(
                    "A state file was found, but it will be ignored as the "
                    "extractor does not advertise the `state` capability",
                )

        return args

    @property
    def config_files(self):  # noqa: ANN201
        """Get the configuration files for this tap."""
        return {
            "config": f"tap.{self.instance_uuid}.config.json",
            "catalog": "tap.properties.json",
            "catalog_cache_key": "tap.properties.cache_key",
            "state": "state.json",
        }

    @property
    def output_files(self):  # noqa: ANN201
        """Get the output files for this tap."""
        return {"output": "tap.out"}

    @hook("before_invoke")
    async def look_up_state_hook(
        self,
        plugin_invoker: PluginInvoker,
        exec_args: tuple[str, ...] = (),
    ) -> None:
        """Look up state before being invoked if in sync mode.

        Args:
            plugin_invoker: the plugin invoker running
            exec_args: the args being passed to the tap

        Returns:
            None
        """
        # Use state only in sync mode (i.e. no args)
        if exec_args:
            return

        with suppress(PluginLacksCapabilityError):
            await self.look_up_state(plugin_invoker)

    async def look_up_state(
        self,
        plugin_invoker: PluginInvoker,
    ) -> None:
        """Look up state, cleaning up and refreshing as needed.

        Args:
            plugin_invoker: the plugin invoker running

        Returns:
            None

        Raises:
            PluginExecutionError: if state could not be found for this plugin
            PluginLacksCapabilityError: if this plugin does not support
                incremental state
        """
        if "state" not in plugin_invoker.capabilities:
            raise PluginLacksCapabilityError(
                f"Extractor '{self.name}' does not support incremental state",  # noqa: EM102
            )

        state_path = plugin_invoker.files["state"]

        with suppress(FileNotFoundError):
            # Delete state left over from different pipeline run for same extractor
            state_path.unlink()
        elt_context = plugin_invoker.context
        if not elt_context or not elt_context.job:
            # Running outside pipeline context: incremental state could not be loaded
            return

        if elt_context.full_refresh:
            logger.info(
                "Performing full refresh, ignoring state left behind by any "
                "previous runs.",
            )
            return

        if custom_state_filename := plugin_invoker.plugin_config_extras["_state"]:
            custom_state_path = plugin_invoker.project.root.joinpath(
                custom_state_filename,
            )

            try:
                shutil.copy(custom_state_path, state_path)
            except FileNotFoundError as err:
                raise PluginExecutionError(
                    f"Could not find state file {custom_state_path}",  # noqa: EM102
                ) from err

            logger.info(f"Found state in {custom_state_filename}")  # noqa: G004
            return

        # the `state.json` is stored in a state backend
        state_service = StateService(
            project=elt_context.project,
            session=elt_context.session,
        )
        try:
            state = state_service.get_state(elt_context.job.job_name)
        except Exception as err:  # pragma: no cover
            logger.error(
                err.args[0],
                state_backend=state_service.state_store_manager.label,
            )
            msg = "Failed to retrieve state"
            raise PluginExecutionError(msg) from err

        if state:
            if state.get(SINGER_STATE_KEY):
                with state_path.open("w") as state_file:
                    json.dump(state.get(SINGER_STATE_KEY), state_file, indent=2)
        else:
            logger.warning("No state was found, complete import.")

    @hook("before_invoke")
    async def discover_catalog_hook(
        self,
        plugin_invoker: PluginInvoker,
        exec_args: tuple[str, ...] = (),
    ) -> None:
        """Discover Singer catalog before invoking tap if in sync mode.

        Args:
            plugin_invoker: The invocation handler of the plugin instance.
            exec_args: List of subcommand/args that we where invoked with.

        Returns:
            None
        """
        # Discover only in sync mode (i.e. no args)
        if exec_args:
            return

        with suppress(PluginLacksCapabilityError):
            await self.discover_catalog(plugin_invoker)

    async def discover_catalog(  # ,
        self,
        plugin_invoker: PluginInvoker,
    ) -> None:
        """Perform catalog discovery.

        Args:
            plugin_invoker: The invocation handler of the plugin instance.

        Returns:
            None

        Raises:
            PluginExecutionError: if discovery could not be performed
        """
        catalog_path = plugin_invoker.files["catalog"]
        catalog_cache_key_path = plugin_invoker.files["catalog_cache_key"]
        elt_context = plugin_invoker.context

        use_catalog_cache = True
        if (
            elt_context
            and elt_context.refresh_catalog
            or not plugin_invoker.plugin_config_extras["_use_cached_catalog"]
        ):
            use_catalog_cache = False

        if catalog_path.exists() and use_catalog_cache:
            with suppress(FileNotFoundError):
                cached_key = catalog_cache_key_path.read_text()
                new_cache_key = self.catalog_cache_key(plugin_invoker)

                if cached_key == new_cache_key:
                    logger.debug("Using cached catalog file")
                    return
            logger.debug("Cached catalog is outdated, running discovery...")

        # We're gonna generate a new catalog, so delete the cache key.
        with suppress(FileNotFoundError):
            catalog_cache_key_path.unlink()

        if custom_catalog_filename := plugin_invoker.plugin_config_extras["_catalog"]:
            custom_catalog_path = plugin_invoker.project.root.joinpath(
                custom_catalog_filename,
            )

            try:
                shutil.copy(custom_catalog_path, catalog_path)
                logger.info(f"Found catalog in {custom_catalog_path}")  # noqa: G004
            except FileNotFoundError as err:
                raise PluginExecutionError(
                    f"Could not find catalog file {custom_catalog_path}",  # noqa: EM102
                ) from err
        else:
            await self.run_discovery(plugin_invoker, catalog_path)

        # test for the result to be a valid catalog
        try:
            with catalog_path.open("r") as catalog_file:
                catalog = json.load(catalog_file)
                Draft4Validator.check_schema(catalog)
        except Exception as err:
            catalog_path.unlink()
            raise PluginExecutionError(
                f"Catalog discovery failed: invalid catalog: {err}",  # noqa: EM102
            ) from err

    async def run_discovery(
        self,
        plugin_invoker: PluginInvoker,
        catalog_path: Path,
    ) -> None:
        """Run tap in discovery mode and store the result.

        Args:
            plugin_invoker: The invocation handler of the plugin instance.
            catalog_path: Where discovery output should be written.

        Raises:
            PluginExecutionError: if state could not be found for this plugin.
            PluginLacksCapabilityError: if this plugin does not support
                incremental state.
            Exception: if any other exception occurs.
        """
        if "discover" not in plugin_invoker.capabilities:
            raise PluginLacksCapabilityError(
                f"Extractor '{self.name}' does not support catalog discovery "  # noqa: EM102
                "(the `discover` capability is not advertised)",
            )
        with StringIO("") as stderr_buff:
            try:
                with catalog_path.open(mode="wb") as catalog:
                    handle = await plugin_invoker.invoke_async(
                        "--discover",
                        stdout=asyncio.subprocess.PIPE,
                        stderr=asyncio.subprocess.PIPE,
                        universal_newlines=False,
                    )

                    invoke_futures = [
                        asyncio.ensure_future(_stream_redirect(handle.stdout, catalog)),
                        asyncio.ensure_future(handle.wait()),
                    ]
                    if logger.isEnabledFor(logging.DEBUG) and handle.stderr:
                        invoke_futures.append(
                            _debug_logging_handler(
                                self.name,
                                plugin_invoker,
                                handle.stderr,
                                stderr_buff,
                            ),
                        )
                    else:
                        invoke_futures.append(
                            asyncio.ensure_future(
                                _stream_redirect(
                                    handle.stderr,
                                    stderr_buff,
                                    write_str=True,
                                ),
                            ),
                        )
                    done, _ = await asyncio.wait(
                        invoke_futures,
                        return_when=asyncio.ALL_COMPLETED,
                    )
                    if failed := [
                        future for future in done if future.exception() is not None
                    ]:
                        failed_future = failed.pop()
                        raise failed_future.exception()  # type: ignore[misc]
                exit_code = handle.returncode
            except Exception:
                catalog_path.unlink()
                raise

            if exit_code != 0:
                catalog_path.unlink()
                stderr_buff.seek(0)
                raise PluginExecutionError(
                    "Catalog discovery failed: command "  # noqa: EM102
                    f"{plugin_invoker.exec_args('--discover')} returned "
                    f"{exit_code} with stderr:\n {stderr_buff.read()}",
                )

    @hook("before_invoke")
    async def apply_catalog_rules_hook(
        self,
        plugin_invoker: PluginInvoker,
        exec_args: tuple[str, ...] = (),
    ) -> None:
        """Apply catalog rules before invoke if in sync mode.

        Args:
            plugin_invoker: the plugin invoker running
            exec_args: the argumnets to pass to the tap

        Returns:
            None
        """
        # Apply only in sync mode (i.e. no args)
        if exec_args:
            return

        with suppress(PluginLacksCapabilityError):
            self.apply_catalog_rules(plugin_invoker, exec_args)

    def apply_catalog_rules(
        self,
        plugin_invoker: PluginInvoker,
        exec_args: tuple[str, ...] = (),  # noqa: ARG002
    ) -> None:
        """Apply Singer catalog and schema rules to discovered catalog.

        Args:
            plugin_invoker: the plugin invoker running
            exec_args: the argumnets to pass to the tap

        Returns:
            None

        Raises:
            PluginLacksCapabilityError: if plugin does not support entity selection
            PluginExecutionError: if catalog rules could not be applied
        """
        if (
            "catalog" not in plugin_invoker.capabilities
            and "properties" not in plugin_invoker.capabilities
        ):
            raise PluginLacksCapabilityError(
                f"Extractor '{self.name}' does not support entity selection "  # noqa: EM102
                "or catalog metadata and schema rules",
            )

        config = plugin_invoker.plugin_config_extras

        schema_rules = []
        metadata_rules = []

        # If a custom catalog is provided, don't apply catalog rules
        if not config["_catalog"]:
            schema_rules.extend(config_schema_rules(config["_schema"]))

            metadata_rules.extend(select_metadata_rules(["!*.*"]))
            metadata_rules.extend(select_metadata_rules(config["_select"]))
            metadata_rules.extend(config_metadata_rules(config["_metadata"]))

        # Always apply select filters (`meltano elt` `--select` and `--exclude` options)
        metadata_rules.extend(select_filter_metadata_rules(config["_select_filter"]))

        if not schema_rules and not metadata_rules:
            return

        catalog_path = plugin_invoker.files["catalog"]
        catalog_cache_key_path = plugin_invoker.files["catalog_cache_key"]

        try:
            with catalog_path.open() as catalog_file:
                catalog = json.load(catalog_file)

            if schema_rules:
                SchemaExecutor(schema_rules).visit(catalog)  # type: ignore[attr-defined]

            if metadata_rules:
                self.warn_property_not_found(metadata_rules, catalog)
                MetadataExecutor(metadata_rules).visit(catalog)  # type: ignore[attr-defined]

            with catalog_path.open("w") as catalog_f:
                json.dump(catalog, catalog_f, indent=2)

            if cache_key := self.catalog_cache_key(plugin_invoker):
                catalog_cache_key_path.write_text(cache_key)
            else:
                with suppress(FileNotFoundError):
                    catalog_cache_key_path.unlink()
        except FileNotFoundError as err:
            raise PluginExecutionError(
                "Applying catalog rules failed: catalog file is missing.",  # noqa: EM101
            ) from err
        except Exception as err:
            catalog_path.unlink()
            raise PluginExecutionError(
                f"Applying catalog rules failed: catalog file is invalid: {err}",  # noqa: EM102
            ) from err

    def catalog_cache_key(self, plugin_invoker):  # noqa: ANN001, ANN201
        """Get a cache key for the catalog.

        Args:
            plugin_invoker: the plugin invoker running

        Returns:
            the cache key for the catalog, if plugin catalog can be cached
        """
        # Treat non-pip plugins as editable/dev-mode plugins and do not cache.
        if plugin_invoker.plugin.pip_url is None:
            return None

        # If the extractor is installed as editable, don't cache because
        # the result of discovery could change at any time.
        if plugin_invoker.plugin.pip_url.startswith("-e"):
            return None

        extras = plugin_invoker.plugin_config_extras

        # If a custom catalog is provided, there's no need to cache
        if extras["_catalog"]:
            return None

        # The catalog should be regenerated, and the catalog cache invalidated,
        # if any settings changed that could affect discovery, or if schema or
        # metadata rules changed.
        # Changes to selection rules and selection filter rules are ignored,
        # since "selected" metadata is reset using the `!*.*` selection rule anyway.
        key_dict = {
            **plugin_invoker.plugin_config,
            "_schema": extras["_schema"],
            "_metadata": extras["_metadata"],
        }

        key_json = json.dumps(key_dict)

        return sha1(key_json.encode()).hexdigest()  # noqa: S324

    @staticmethod
    @lru_cache
    def _warn_missing_stream(stream_id: str) -> None:
        logger.warning(
            "Stream `%s` was not found in the catalog",
            stream_id,
        )

    @staticmethod
    @lru_cache
    def _warn_missing_property(stream_id: str, breadcrumb: tuple[str, ...]) -> None:
        logger.warning(
            "Property `%s` was not found in the schema of stream `%s`",
            ".".join(breadcrumb[1:]),
            stream_id,
        )

    def warn_property_not_found(
        self,
        rules: list[MetadataRule],
        catalog: CatalogDict,
    ) -> None:
        """Validate MetadataRules conforms to discovered Catalog.

        Validate MetadataRules against the tap's discovered Catalog & emit
        warnings for each rule that does not match any property in catalog.

        The method is intentionally written in a defensive manner so as not to
        raise exceptions. e.g.
            * use dict.get rather than []-accessor
            * filter list comprehensions to remove some elements that should
              not be used for validation

        Args:
            rules: List of `MetadataRule`
            catalog: Discovered Source Catalog
        """
        stream_dict = {
            stream.get("tap_stream_id"): stream
            for stream in catalog.get("streams", [])
            if isinstance(stream, dict)
        }

        def is_not_star(x):  # noqa: ANN001, ANN202
            return "*" not in x

        def dict_get(dictionary, key):  # noqa: ANN001, ANN202
            return dictionary.get(key, {})

        for rule in rules:
            if isinstance(rule.tap_stream_id, list) or "*" in rule.tap_stream_id:
                continue
            if not (s := stream_dict.get(rule.tap_stream_id)):
                self._warn_missing_stream(rule.tap_stream_id)
                continue
            path = tuple(takewhile(is_not_star, rule.breadcrumb))
            if len(path) <= 1:
                continue
            if not reduce(dict_get, path, s.get("schema", {})):
                self._warn_missing_property(rule.tap_stream_id, tuple(rule.breadcrumb))



from __future__ import annotations

import asyncio
import json
import logging
import subprocess
import sys
import typing as t
from contextlib import contextmanager, suppress

import pytest
from mock import AsyncMock, mock

from meltano.core.job import Job, Payload
from meltano.core.plugin import PluginType
from meltano.core.plugin.error import PluginExecutionError
from meltano.core.plugin.singer import SingerTap
from meltano.core.plugin.singer.catalog import (
    CatalogDict,
    ListSelectedExecutor,
    select_metadata_rules,
)
from meltano.core.state_service import InvalidJobStateError, StateService


class CatalogFixture:
    empty_stream: t.ClassVar[CatalogDict] = {"streams": []}
    empty_properties: t.ClassVar[CatalogDict] = {"streams": [{"tap_stream_id": "foo"}]}
    regular_stream: t.ClassVar[CatalogDict] = {
        "streams": [
            {
                "tap_stream_id": "foo",
                "schema": {
                    "properties": {
                        "bar": {"type": "string"},
                        "attribute": {
                            "type": "object",
                            "properties": {
                                "name": {"type": "string"},
                            },
                        },
                    },
                },
            },
        ],
    }


class TestSingerTap:
    @pytest.fixture(scope="class")
    def subject(self, project_add_service):
        return project_add_service.add(PluginType.EXTRACTORS, "tap-mock")

    @pytest.mark.order(0)
    @pytest.mark.asyncio()
    async def test_exec_args(self, subject, session, plugin_invoker_factory) -> None:
        invoker = plugin_invoker_factory(subject)
        async with invoker.prepared(session):
            assert subject.exec_args(invoker) == ["--config", invoker.files["config"]]

            # when `catalog` has data
            invoker.files["catalog"].open("w").write("...")
            assert subject.exec_args(invoker) == [
                "--config",
                invoker.files["config"],
                "--catalog",
                invoker.files["catalog"],
            ]

            # when `state` has data
            invoker.files["state"].open("w").write("...")
            assert subject.exec_args(invoker) == [
                "--config",
                invoker.files["config"],
                "--catalog",
                invoker.files["catalog"],
                "--state",
                invoker.files["state"],
            ]

    @pytest.mark.asyncio()
    async def test_cleanup(self, subject, session, plugin_invoker_factory) -> None:
        invoker = plugin_invoker_factory(subject)
        async with invoker.prepared(session):
            assert invoker.files["config"].exists()

        assert not invoker.files["config"].exists()

    @pytest.mark.asyncio()
    async def test_look_up_state(
        self,
        subject,
        project,
        session,
        plugin_invoker_factory,
        elt_context_builder,
        monkeypatch,
    ) -> None:
        job = Job(job_name="pytest_test_runner")
        elt_context = (
            elt_context_builder.with_session(session)
            .with_extractor(subject.name)
            .with_job(job)
            .context()
        )

        invoker = plugin_invoker_factory(subject, context=elt_context)
        state_service = StateService(session=session)

        @contextmanager
        def create_job():
            new_job = Job(job_name=job.job_name)
            new_job.start()
            yield new_job
            new_job.save(session)
            if new_job.payload and not new_job.is_running():
                with suppress(InvalidJobStateError):
                    state_service.add_state(
                        new_job.job_name,
                        json.dumps(new_job.payload),
                        new_job.payload_flags,
                    )

        async def assert_state(state) -> None:
            async with invoker.prepared(session):
                await subject.look_up_state(invoker)

            if state:
                assert invoker.files["state"].exists()
                assert json.load(invoker.files["state"].open()) == state
            else:
                assert not invoker.files["state"].exists()

        # No state by default
        await assert_state(None)

        # Running jobs with state are not considered
        with create_job() as job:
            job.payload["singer_state"] = {"success": True}
            job.payload_flags = Payload.STATE

        await assert_state(None)

        # Successful jobs without state are not considered
        with create_job() as job:
            job.success()
        await assert_state(None)

        # Successful jobs with incomplete state are considered
        with create_job() as job:
            job.payload["singer_state"] = {"incomplete_success": True}
            job.payload_flags = Payload.INCOMPLETE_STATE
            job.success()

        await assert_state({"incomplete_success": True})

        # Successful jobs with state are considered
        with create_job() as job:
            job.payload["singer_state"] = {"success": True}
            job.payload_flags = Payload.STATE
            job.success()
        await assert_state({"success": True})

        # Failed jobs without state are not considered
        with create_job() as job:
            job.fail("Whoops")

        await assert_state({"success": True})

        # Failed jobs with state are considered
        with create_job() as job:
            job.payload["singer_state"] = {"failed": True}
            job.payload_flags = Payload.STATE
            job.fail("Whoops")

        await assert_state({"failed": True})

        # Successful jobs with incomplete state are considered
        with create_job() as job:
            job.payload["singer_state"] = {"success": True}
            job.payload_flags = Payload.INCOMPLETE_STATE
            job.success()

        # Incomplete state is merged into complete state
        await assert_state({"failed": True, "success": True})

        # Failed jobs with incomplete state are considered
        with create_job() as job:
            job.payload["singer_state"] = {"failed_again": True}
            job.payload_flags = Payload.INCOMPLETE_STATE
            job.fail("Whoops")

        # Incomplete state is merged into complete state
        await assert_state({"failed": True, "success": True, "failed_again": True})

        # Custom state takes precedence
        custom_state_filename = "custom_state.json"
        custom_state_path = project.root.joinpath(custom_state_filename)
        custom_state_path.write_text('{"custom": true}')

        monkeypatch.setitem(
            invoker.settings_service.config_override,
            "_state",
            custom_state_filename,
        )

        await assert_state({"custom": True})

        # With a full refresh, no state is considered
        elt_context.full_refresh = True
        await assert_state(None)

    @pytest.mark.order(1)
    @pytest.mark.asyncio()
    async def test_discover_catalog(
        self,
        session,
        plugin_invoker_factory,
        subject,
        monkeypatch,
    ) -> None:
        invoker = plugin_invoker_factory(subject)

        catalog_path = invoker.files["catalog"]
        catalog_cache_key_path = invoker.files["catalog_cache_key"]

        def mock_discovery(*args, **kwargs):  # noqa: ARG001
            future = asyncio.Future()
            future.set_result(catalog_path.open("w").write('{"discovered": true}'))
            return future

        async with invoker.prepared(session):
            with mock.patch.object(
                SingerTap,
                "run_discovery",
                side_effect=mock_discovery,
            ) as mocked_run_discovery:
                await subject.discover_catalog(invoker)

                assert mocked_run_discovery.called
                assert json.loads(catalog_path.read_text()) == {"discovered": True}
                assert not catalog_cache_key_path.exists()

                # If there is no cache key, discovery is invoked again
                mocked_run_discovery.reset_mock()
                await subject.discover_catalog(invoker)

                assert json.loads(catalog_path.read_text()) == {"discovered": True}
                assert not catalog_cache_key_path.exists()

                # Apply catalog rules to store the cache key
                subject.apply_catalog_rules(invoker)
                assert catalog_cache_key_path.exists()

                # If the cache key hasn't changed, discovery isn't invoked again
                mocked_run_discovery.reset_mock()
                await subject.discover_catalog(invoker)

                mocked_run_discovery.assert_not_called()
                assert json.loads(catalog_path.read_text()) == {"discovered": True}
                assert catalog_cache_key_path.exists()

                # If the cache key no longer matches, discovery is invoked again
                mocked_run_discovery.reset_mock()
                catalog_cache_key_path.write_text("bogus")
                await subject.discover_catalog(invoker)

                assert mocked_run_discovery.called
                assert json.loads(catalog_path.read_text()) == {"discovered": True}
                assert not catalog_cache_key_path.exists()

                # Apply catalog rules to store the cache key again
                subject.apply_catalog_rules(invoker)
                assert catalog_cache_key_path.exists()

        monkeypatch.setitem(
            invoker.settings_service.config_override,
            "_use_cached_catalog",
            value=False,
        )

        async with invoker.prepared(session):
            with mock.patch.object(
                SingerTap,
                "run_discovery",
                side_effect=mock_discovery,
            ) as mocked_run_discovery:
                assert catalog_cache_key_path.exists()

                # with _use_cached_catalog = false, discovery should be invoked
                # again even with a stored cache key.
                mocked_run_discovery.reset_mock()
                await subject.discover_catalog(invoker)

                mocked_run_discovery.assert_called_once()
                assert json.loads(catalog_path.read_text()) == {"discovered": True}
                assert not catalog_cache_key_path.exists()

    @pytest.mark.asyncio()
    async def test_discover_catalog_custom(
        self,
        project,
        session,
        plugin_invoker_factory,
        subject,
        monkeypatch,
    ) -> None:
        invoker = plugin_invoker_factory(subject)

        custom_catalog_filename = "custom_catalog.json"
        custom_catalog_path = project.root.joinpath(custom_catalog_filename)
        custom_catalog_path.write_text('{"custom": true}')

        monkeypatch.setitem(
            invoker.settings_service.config_override,
            "_catalog",
            custom_catalog_filename,
        )

        async with invoker.prepared(session):
            await subject.discover_catalog(invoker)

        assert invoker.files["catalog"].read_text() == '{"custom": true}'

    @pytest.mark.asyncio()
    async def test_apply_select(
        self,
        session,
        plugin_invoker_factory,
        subject,
        monkeypatch,
    ) -> None:
        invoker = plugin_invoker_factory(subject)

        catalog_path = invoker.files["catalog"]

        def reset_catalog() -> None:
            catalog_path.open("w").write('{"rules": []}')

        def assert_rules(*rules) -> None:
            with catalog_path.open() as catalog_file:
                catalog = json.load(catalog_file)

            assert catalog["rules"] == list(rules)

        def mock_metadata_executor(rules):
            def visit(catalog) -> None:
                for rule in rules:
                    catalog["rules"].append(
                        [rule.tap_stream_id, rule.breadcrumb, rule.key, rule.value],
                    )

            return mock.Mock(visit=visit)

        with mock.patch(
            "meltano.core.plugin.singer.tap.MetadataExecutor",
            side_effect=mock_metadata_executor,
        ):
            reset_catalog()

            async with invoker.prepared(session):
                subject.apply_catalog_rules(invoker)

            # When `select` isn't set in meltano.yml or the plugin def, select all
            assert_rules(
                ["*", [], "selected", False],
                ["*", ["properties", "*"], "selected", False],
                ["*", [], "selected", True],
                ["*", ["properties", "*"], "selected", True],
            )

            reset_catalog()

            # Pretend `select` is set in the plugin definition
            monkeypatch.setitem(
                invoker.plugin.parent._variant.extras,
                "select",
                ["UniqueEntitiesName.name"],
            )
            invoker.settings_service._setting_defs = None
            async with invoker.prepared(session):
                subject.apply_catalog_rules(invoker)

            # When `select` is set in the plugin definition, use the selection
            assert_rules(
                ["*", [], "selected", False],
                ["*", ["properties", "*"], "selected", False],
                ["UniqueEntitiesName", [], "selected", True],
                ["UniqueEntitiesName", ["properties", "name"], "selected", True],
            )

            reset_catalog()

            # Pretend `select` is set in meltano.yml
            monkeypatch.setitem(
                invoker.plugin.extras,
                "select",
                ["UniqueEntitiesName.code"],
            )

            async with invoker.prepared(session):
                subject.apply_catalog_rules(invoker)

            # `select` set in meltano.yml takes precedence over the plugin definition
            assert_rules(
                ["*", [], "selected", False],
                ["*", ["properties", "*"], "selected", False],
                ["UniqueEntitiesName", [], "selected", True],
                ["UniqueEntitiesName", ["properties", "code"], "selected", True],
            )

    @pytest.mark.asyncio()
    async def test_apply_catalog_rules(
        self,
        session,
        plugin_invoker_factory,
        subject,
        monkeypatch,
    ) -> None:
        invoker = plugin_invoker_factory(subject)

        catalog_path = invoker.files["catalog"]
        catalog_cache_key_path = invoker.files["catalog_cache_key"]

        def reset_catalog() -> None:
            catalog_path.open("w").write('{"rules": []}')

        def assert_rules(*rules) -> None:
            with catalog_path.open() as catalog_file:
                catalog = json.load(catalog_file)

            assert catalog["rules"] == list(rules)

        def mock_metadata_executor(rules):
            def visit(catalog) -> None:
                for rule in rules:
                    rule_list = [
                        rule.tap_stream_id,
                        rule.breadcrumb,
                        rule.key,
                        rule.value,
                    ]
                    if rule.negated:
                        rule_list.append({"negated": True})
                    catalog["rules"].append(rule_list)

            return mock.Mock(visit=visit)

        def mock_schema_executor(rules):
            def visit(catalog) -> None:
                for rule in rules:
                    catalog["rules"].append(
                        [rule.tap_stream_id, rule.breadcrumb, rule.payload],
                    )

            return mock.Mock(visit=visit)

        with mock.patch(
            "meltano.core.plugin.singer.tap.MetadataExecutor",
            side_effect=mock_metadata_executor,
        ), mock.patch(
            "meltano.core.plugin.singer.tap.SchemaExecutor",
            side_effect=mock_schema_executor,
        ):
            reset_catalog()

            config = {
                "_select": ["UniqueEntitiesName.code"],
                "_metadata": {
                    "UniqueEntitiesName": {"replication-key": "created_at"},
                    "UniqueEntitiesName.created_at": {"is-replication-key": True},
                },
                "metadata.UniqueEntitiesName.properties.payload.properties.hash.custom-metadata": "custom-value",  # noqa: E501
                "_schema": {
                    "UniqueEntitiesName": {
                        "code": {"anyOf": [{"type": "string"}, {"type": "null"}]},
                    },
                    "UniqueEntitiesName.payload.type": "object",
                    "UniqueEntitiesName.payload.properties": {
                        "content": {"type": ["string", "null"]},
                        "hash": {"type": "string"},
                    },
                },
                "_select_filter": [
                    "UniqueEntitiesName",
                    "!OtherEntitiesName",
                    "SelectAnother",
                    "!ExcludeAnother",
                ],
            }

            # Pretend `config` is set in meltano.yml
            monkeypatch.setattr(invoker.plugin, "config", config)

            async with invoker.prepared(session):
                subject.apply_catalog_rules(invoker)

                cache_key = invoker.plugin.catalog_cache_key(invoker)

            # Stores the cache key
            assert catalog_cache_key_path.read_text() == cache_key

            assert_rules(
                # Schema rules
                [
                    "UniqueEntitiesName",
                    ["properties", "code"],
                    {"anyOf": [{"type": "string"}, {"type": "null"}]},
                ],
                [
                    "UniqueEntitiesName",
                    ["properties", "payload"],
                    {
                        "type": "object",
                        "properties": {
                            "content": {"type": ["string", "null"]},
                            "hash": {"type": "string"},
                        },
                    },
                ],
                # Clean slate selection metadata rules
                ["*", [], "selected", False],
                ["*", ["properties", "*"], "selected", False],
                # Selection metadata rules
                ["UniqueEntitiesName", [], "selected", True],
                ["UniqueEntitiesName", ["properties", "code"], "selected", True],
                # Metadata rules
                ["UniqueEntitiesName", [], "replication-key", "created_at"],
                [
                    "UniqueEntitiesName",
                    ["properties", "created_at"],
                    "is-replication-key",
                    True,
                ],
                [
                    "UniqueEntitiesName",
                    ["properties", "payload", "properties", "hash"],
                    "custom-metadata",
                    "custom-value",
                ],
                # Selection filter metadata rules
                [
                    ["UniqueEntitiesName", "SelectAnother"],
                    [],
                    "selected",
                    False,
                    {"negated": True},
                ],
                [["OtherEntitiesName", "ExcludeAnother"], [], "selected", False],
            )

            reset_catalog()

            # If a custom catalog is provided, only selection filters are applied
            config_override = invoker.settings_service.config_override
            monkeypatch.setitem(config_override, "_catalog", "custom_catalog.json")

            async with invoker.prepared(session):
                subject.apply_catalog_rules(invoker)

                cache_key = invoker.plugin.catalog_cache_key(invoker)

            assert_rules(
                # Selection filter metadata rules
                [
                    ["UniqueEntitiesName", "SelectAnother"],
                    [],
                    "selected",
                    False,
                    {"negated": True},
                ],
                [["OtherEntitiesName", "ExcludeAnother"], [], "selected", False],
            )

            # Doesn't store a cache key when a custom catalog is provided
            assert not catalog_cache_key_path.exists()
            assert cache_key is None

    @pytest.mark.asyncio()
    async def test_apply_catalog_rules_select_filter(
        self,
        session,
        plugin_invoker_factory,
        subject,
        monkeypatch,
    ) -> None:
        invoker = plugin_invoker_factory(subject)

        stream_data = {
            "schema": {
                "type": "object",
                "properties": {
                    "one": {"type": "string"},
                    "two": {"type": "string"},
                    "three": {"type": "string"},
                },
            },
            "metadata": [
                {"breadcrumb": [], "metadata": {"inclusion": "available"}},
                {
                    "breadcrumb": ["properties", "one"],
                    "metadata": {"inclusion": "automatic"},
                },
                {
                    "breadcrumb": ["properties", "two"],
                    "metadata": {"inclusion": "available"},
                },
                {
                    "breadcrumb": ["properties", "three"],
                    "metadata": {"inclusion": "available"},
                },
            ],
        }

        base_catalog = {
            "streams": [
                {"tap_stream_id": "one", **stream_data},
                {"tap_stream_id": "two", **stream_data},
                {"tap_stream_id": "three", **stream_data},
                {"tap_stream_id": "four", **stream_data},
                {"tap_stream_id": "five", **stream_data},
            ],
        }

        async def selected_properties():
            catalog_path = invoker.files["catalog"]

            with catalog_path.open("w") as catalog_file:
                json.dump(base_catalog, catalog_file)

            async with invoker.prepared(session):
                subject.apply_catalog_rules(invoker, [])

            with catalog_path.open() as catalog_file:
                catalog = json.load(catalog_file)

            lister = ListSelectedExecutor()
            lister.visit(catalog)

            return lister.selected_properties

        config_override = invoker.settings_service.config_override

        monkeypatch.setitem(
            config_override,
            "_select",
            ["one.one", "three.three", "five.*"],
        )

        # `one` is always included because it has `inclusion: automatic`
        assert await selected_properties() == {
            "one": {"one"},
            "three": {"one", "three"},
            "five": {"one", "two", "three"},
        }

        # Simple inclusion
        monkeypatch.setitem(config_override, "_select_filter", ["three"])
        assert await selected_properties() == {"three": {"one", "three"}}

        # Simple exclusion
        monkeypatch.setitem(config_override, "_select_filter", ["!three"])
        assert await selected_properties() == {
            "one": {"one"},
            "five": {"one", "two", "three"},
        }

        # Wildcard inclusion
        monkeypatch.setitem(config_override, "_select_filter", ["t*"])
        assert await selected_properties() == {"three": {"one", "three"}}

        # Wildcard exclusion
        monkeypatch.setitem(config_override, "_select_filter", ["!t*"])
        assert await selected_properties() == {
            "one": {"one"},
            "five": {"one", "two", "three"},
        }

        # Multiple inclusion
        monkeypatch.setitem(config_override, "_select_filter", ["three", "five"])
        assert await selected_properties() == {
            "three": {"one", "three"},
            "five": {"one", "two", "three"},
        }

        # Multiple exclusion
        monkeypatch.setitem(config_override, "_select_filter", ["!three", "!five"])
        assert await selected_properties() == {"one": {"one"}}

        # Multiple wildcard inclusion
        monkeypatch.setitem(config_override, "_select_filter", ["t*", "f*"])
        assert await selected_properties() == {
            "three": {"one", "three"},
            "five": {"one", "two", "three"},
        }

        # Multiple wildcard exclusion
        monkeypatch.setitem(config_override, "_select_filter", ["!t*", "!f*"])
        assert await selected_properties() == {"one": {"one"}}

        # Mixed inclusion and exclusion
        monkeypatch.setitem(config_override, "_select_filter", ["*e", "!*ee"])
        assert await selected_properties() == {
            "one": {"one"},
            "five": {"one", "two", "three"},
        }

    @pytest.mark.asyncio()
    async def test_apply_catalog_rules_invalid(
        self,
        session,
        plugin_invoker_factory,
        subject,
    ) -> None:
        invoker = plugin_invoker_factory(subject)
        async with invoker.prepared(session):
            invoker.files["catalog"].open("w").write("this is invalid json")

            with pytest.raises(PluginExecutionError, match=r"invalid"):
                subject.apply_catalog_rules(invoker, [])

    @pytest.mark.asyncio()
    async def test_catalog_cache_key(
        self,
        session,
        plugin_invoker_factory,
        subject,
        monkeypatch,
    ) -> None:
        invoker = plugin_invoker_factory(subject)
        config_override = invoker.settings_service.config_override

        async def cache_key():
            async with invoker.prepared(session):
                return subject.catalog_cache_key(invoker)

        # No key if _catalog is set
        monkeypatch.setitem(config_override, "_catalog", "catalog.json")
        assert await cache_key() is None

        # Key is set if _catalog is not set
        monkeypatch.setitem(config_override, "_catalog", None)
        key = await cache_key()
        assert key is not None

        # Key doesn't change if nothing has changed
        assert await cache_key() == key

        # Key changes if config changes
        monkeypatch.setitem(config_override, "test", "foo")

        new_key = await cache_key()
        assert new_key != key
        key = new_key

        # Key changes if _schema changes
        monkeypatch.setitem(
            config_override,
            "_schema",
            {"stream": {"property": {"type": "string"}}},
        )

        new_key = await cache_key()
        assert new_key != key
        key = new_key

        # Key changes if _metadata changes
        monkeypatch.setitem(
            config_override,
            "_metadata",
            {"stream": {"property": {"is-replication-key": True}}},
        )

        new_key = await cache_key()
        assert new_key != key
        key = new_key

        # Key does not change if _select changes
        monkeypatch.setitem(config_override, "_select", ["stream"])
        assert await cache_key() == key

        # Key does not change if _select_filter changes
        monkeypatch.setitem(config_override, "_select_filter", ["stream"])
        assert await cache_key() == key

        # No key if pip_url is editable
        monkeypatch.setattr(invoker.plugin, "pip_url", "-e local")
        assert await cache_key() is None

    @pytest.mark.asyncio()
    @pytest.mark.usefixtures("session", "elt_context_builder")
    async def test_run_discovery(
        self,
        plugin_invoker_factory,
        subject,
    ) -> None:
        process_mock = mock.Mock()
        process_mock.name = subject.name
        process_mock.wait = AsyncMock(return_value=0)
        process_mock.returncode = 0
        process_mock.sterr.at_eof.side_effect = (
            True  # no output so return eof immediately
        )

        # First check needs to be false so loop starts read, after 1 line,
        # we'll return true
        process_mock.stdout.at_eof.side_effect = (False, True)
        process_mock.stdout.readline = AsyncMock(return_value=b'{"discovered": true}\n')

        invoke_async = AsyncMock(return_value=process_mock)
        invoker = plugin_invoker_factory(subject)
        invoker.invoke_async = invoke_async
        catalog_path = invoker.files["catalog"]

        await subject.run_discovery(invoker, catalog_path)
        assert invoke_async.call_args[0] == ("--discover",)

        with catalog_path.open("r") as catalog_file:
            resp = json.load(catalog_file)
            assert resp["discovered"]

    @pytest.mark.asyncio()
    @pytest.mark.usefixtures("session", "elt_context_builder")
    async def test_run_discovery_failure(
        self,
        plugin_invoker_factory,
        subject,
    ) -> None:
        process_mock = mock.Mock()
        process_mock.name = subject.name
        process_mock.wait = AsyncMock(return_value=1)
        process_mock.returncode = 1
        process_mock.stderr.at_eof.side_effect = (False, True)
        process_mock.stderr.readline = AsyncMock(return_value=b"stderr mock output")
        process_mock.stdout.at_eof.side_effect = (True, True)
        process_mock.stdout.readline = AsyncMock(return_value=b"")

        invoker = plugin_invoker_factory(subject)
        invoker.invoke_async = AsyncMock(return_value=process_mock)
        catalog_path = invoker.files["catalog"]

        with pytest.raises(PluginExecutionError, match="returned 1"):
            await subject.run_discovery(invoker, catalog_path)

        assert not catalog_path.exists(), "Catalog should not be present."

    @pytest.mark.asyncio()
    @pytest.mark.usefixtures("session", "elt_context_builder")
    async def test_run_discovery_stderr_output(
        self,
        plugin_invoker_factory,
        subject,
    ) -> None:
        process_mock = mock.Mock()
        process_mock.name = subject.name
        # we need to exit successfully to not trigger error handling
        process_mock.wait = AsyncMock(return_value=0)
        process_mock.returncode = 0
        process_mock.stderr.at_eof.side_effect = (False, True)
        process_mock.stderr.readline = AsyncMock(return_value=b"stderr mock output")
        process_mock.stdout.at_eof.side_effect = (True, True)
        process_mock.stdout.readline = AsyncMock(return_value=b"")

        invoker = plugin_invoker_factory(subject)
        invoker.invoke_async = AsyncMock(return_value=process_mock)
        catalog_path = invoker.files["catalog"]

        with mock.patch(
            "meltano.core.plugin.singer.tap.logger.isEnabledFor",
            return_value=False,
        ), mock.patch("meltano.core.plugin.singer.tap._stream_redirect") as stream_mock:
            await subject.run_discovery(invoker, catalog_path)
            assert stream_mock.call_count == 2

        with mock.patch(
            "meltano.core.plugin.singer.tap.logger.isEnabledFor",
            return_value=True,
        ), mock.patch(
            "meltano.core.plugin.singer.tap._stream_redirect",
        ) as stream_mock2:
            await subject.run_discovery(invoker, catalog_path)
            assert stream_mock2.call_count == 2

        # ensure stderr is redirected to devnull if we don't need it
        discovery_logger = logging.getLogger("meltano.core.plugin.singer.tap")  # noqa: TID251
        original_level = discovery_logger.getEffectiveLevel()
        discovery_logger.setLevel(logging.INFO)
        with mock.patch(
            "meltano.core.plugin.singer.tap.logger.isEnabledFor",
            return_value=True,
        ), mock.patch(
            "meltano.core.plugin.singer.tap._stream_redirect",
        ) as stream_mock3:
            await subject.run_discovery(invoker, catalog_path)

            assert stream_mock3.call_count == 2
            call_kwargs = invoker.invoke_async.call_args_list[0][1]
            assert call_kwargs.get("stderr") is subprocess.PIPE

        discovery_logger.setLevel(original_level)

    @pytest.mark.asyncio()
    @pytest.mark.usefixtures("session", "elt_context_builder")
    async def test_run_discovery_handle_io_exceptions(
        self,
        plugin_invoker_factory,
        subject,
    ) -> None:
        process_mock = mock.Mock()
        process_mock.name = subject.name
        process_mock.wait = AsyncMock(return_value=0)
        process_mock.returncode = 0
        process_mock.stderr.at_eof.side_effect = (False, True)
        process_mock.stderr.readline = AsyncMock(return_value=b"stderr mock output")
        process_mock.stdout.at_eof.side_effect = (False, True)
        process_mock.stdout.readline.side_effect = Exception("mock readline exception")

        invoker = plugin_invoker_factory(subject)
        invoker.invoke_async = AsyncMock(return_value=process_mock)
        catalog_path = invoker.files["catalog"]

        with pytest.raises(Exception, match="mock readline exception"):
            await subject.run_discovery(invoker, catalog_path)

        assert not catalog_path.exists(), "Catalog should not be present."

    @pytest.mark.asyncio()
    @pytest.mark.usefixtures("session", "elt_context_builder")
    async def test_run_discovery_utf8_output(
        self,
        plugin_invoker_factory,
        subject,
    ) -> None:
        process_mock = mock.Mock()
        process_mock.name = subject.name
        # we need to exit successfully to not trigger error handling
        process_mock.wait = AsyncMock(return_value=0)
        process_mock.returncode = 0
        process_mock.stderr.at_eof.side_effect = (False, True)
        test_string = "Hello world, Καλημέρα κόσμε, コンニチハ".encode()
        process_mock.stderr.readline = AsyncMock(return_value=test_string)
        process_mock.stdout.at_eof.side_effect = (True, True)
        process_mock.stdout.readline = AsyncMock(return_value=b"")

        invoker = plugin_invoker_factory(subject)
        invoker.invoke_async = AsyncMock(return_value=process_mock)
        catalog_path = invoker.files["catalog"]

        assert sys.getdefaultencoding() == "utf-8"

        with mock.patch(
            "meltano.core.plugin.singer.tap.logger.isEnabledFor",
            return_value=True,
        ):
            await subject.run_discovery(invoker, catalog_path)

    @pytest.mark.asyncio()
    @pytest.mark.usefixtures("use_test_log_config")
    @pytest.mark.parametrize(
        ("rule_pattern", "catalog", "message"),
        (
            pytest.param(
                "bar.*",
                CatalogFixture.empty_properties,
                "Stream `bar` was not found",
                id="stream_not_found",
            ),
            pytest.param(
                "foo.*",
                CatalogFixture.empty_stream,
                "Stream `foo` was not found",
                id="no_streams_exist",
            ),
            pytest.param(
                "foo.bar",
                CatalogFixture.empty_properties,
                "Property `bar` was not found in the schema of stream `foo`",
                id="no_properties_exist",
            ),
            pytest.param(
                "foo.spam",
                CatalogFixture.regular_stream,
                "Property `spam` was not found in the schema of stream `foo`",
                id="property_not_found",
            ),
            pytest.param(
                "foo.bar",
                CatalogFixture.regular_stream,
                None,
                id="property_is_found",
            ),
            pytest.param(
                "*.bar",
                CatalogFixture.empty_stream,
                None,
                id="stream_is_wildcard",
            ),
            pytest.param(
                "zoo.*",
                CatalogFixture.empty_stream,
                "Stream `zoo` was not found",
                id="stream_not_found_property_is_wildcard",
            ),
            pytest.param(
                "foo.*",
                CatalogFixture.regular_stream,
                None,
                id="property_is_wildcard",
            ),
            pytest.param(
                "foo.attribute.name",
                CatalogFixture.regular_stream,
                None,
                id="nested_property_is_found",
            ),
            pytest.param(
                "foo.attribute.*",
                CatalogFixture.regular_stream,
                None,
                id="nested_property_is_wildcard",
            ),
            pytest.param(
                "foo.attribute.email",
                CatalogFixture.regular_stream,
                "Property `attribute.properties.email` was not "
                "found in the schema of stream `foo`",
                id="nested_property_not_found",
            ),
            pytest.param(
                "foo*.*",
                CatalogFixture.regular_stream,
                None,
                id="stream_has_wildcard",
            ),
            pytest.param(
                "foo.ATTRIBUTES.em*",
                CatalogFixture.regular_stream,
                "Property `ATTRIBUTES.properties.em*` was not "
                "found in the schema of stream `foo`",
                id="property_not_found-sub_property_has_wildcard",
            ),
        ),
    )
    async def test_warn_property_not_found(
        self,
        subject: SingerTap,
        caplog: pytest.LogCaptureFixture,
        capsys: pytest.CaptureFixture,
        rule_pattern: str,
        catalog: CatalogDict,
        message: str | None,
    ) -> None:
        """
        Warning messages should be emitted when a MetadataRule doesn't match.

        _A quirk seems to be that if this test is invoked on its own, the
        output is captured in the fixture `pytest.capsys`, however if pytest
        runs globally then the output is captured in `pytest.caplog`. Not sure
        why, hoping this message finds a developer more knowledgable than I._

        For now, I'm adding up the output of both capsys & caplog and testing
        expected messages are captured somewhere. As in, both of these would
        work.

        ```
        poetry run pytest
        poetry run pytest tests/meltano/core/plugin/singer/test_tap.py::TestSingerTap::test_warn_property_not_found
        ```
        """  # noqa: D212, E501
        rules = select_metadata_rules([rule_pattern])
        subject.warn_property_not_found(rules, catalog)
        sysout, syserr = capsys.readouterr()
        # If the warning is emitted, it should be in the output
        if message is not None:
            assert message in (caplog.text + sysout + syserr)
        else:
            assert len(caplog.records) == 0
            assert syserr + sysout == ""



from __future__ import annotations

import json
import shutil
import typing as t
from copy import deepcopy

import pytest

from meltano.core.plugin import BasePlugin, PluginType
from meltano.core.plugin.error import PluginNotFoundError, PluginParentNotFoundError
from meltano.core.plugin.project_plugin import ProjectPlugin
from meltano.core.project_plugins_service import (
    DefinitionSource,
    PluginDefinitionNotFoundError,
)

if t.TYPE_CHECKING:
    from meltano.core.locked_definition_service import LockedDefinitionService
    from meltano.core.project import Project


@pytest.fixture()
def modified_lockfile(project: Project):
    lockfile_path = project.plugin_lock_path(
        PluginType.EXTRACTORS,
        "tap-mock",
        variant_name="meltano",
    )
    with lockfile_path.open() as lockfile:
        original_contents = json.load(lockfile)

    new_contents = deepcopy(original_contents)
    new_contents["settings"].append({"name": "foo"})

    with lockfile_path.open("w") as lockfile:
        json.dump(new_contents, lockfile)

    yield

    with lockfile_path.open("w") as lockfile:
        json.dump(original_contents, lockfile)


class TestProjectPluginsService:
    @pytest.fixture(autouse=True)
    def setup(self, project: Project, tap) -> None:
        project.plugins.lock_service.save(tap, exists_ok=True)
        project.plugins._prefer_source = DefinitionSource.ANY

    @pytest.mark.order(0)
    def test_plugins(self, project) -> None:
        assert all(
            isinstance(plugin.parent, BasePlugin)
            for plugin in project.plugins.plugins()
        )

    def test_get_plugin(
        self,
        project,
        tap,
        alternative_tap,
        inherited_tap,
        alternative_target,
    ) -> None:
        # name="tap-mock", variant="meltano"
        plugin = project.plugins.get_plugin(tap)
        assert plugin.type == PluginType.EXTRACTORS
        assert plugin.name == "tap-mock"
        assert plugin.variant == "meltano"
        assert plugin.parent

        # name="tap-mock--singer-io", inherit_from="tap-mock", variant="singer-io"
        alternative_plugin = project.plugins.get_plugin(alternative_tap)
        assert alternative_plugin.type == PluginType.EXTRACTORS
        assert alternative_plugin.name == "tap-mock--singer-io"
        assert alternative_plugin.inherit_from == "tap-mock"
        assert alternative_plugin.variant == "singer-io"
        assert plugin.parent

        # name="tap-mock-inherited", inherit_from="tap-mock"
        inherited_plugin = project.plugins.get_plugin(inherited_tap)
        assert inherited_plugin.type == PluginType.EXTRACTORS
        assert inherited_plugin.name == "tap-mock-inherited"
        assert inherited_plugin.inherit_from == "tap-mock"
        assert plugin.parent

        # name="target-mock-alternative", inherit_from="target-mock"
        alternative_plugin = project.plugins.get_plugin(alternative_target)
        assert alternative_plugin.type == PluginType.LOADERS
        assert alternative_plugin.name == "target-mock-alternative"
        assert alternative_plugin.inherit_from == "target-mock"
        assert plugin.parent

        # Results are cached
        assert project.plugins.get_plugin(tap) is plugin

        project.refresh()
        assert project.plugins.get_plugin(tap) is not plugin

    @pytest.mark.order(2)
    @pytest.mark.usefixtures("modified_lockfile")
    def test_get_parent_from_lockfile(
        self,
        project: Project,
        tap: ProjectPlugin,
        locked_definition_service: LockedDefinitionService,
    ) -> None:
        expected = locked_definition_service.find_base_plugin(
            plugin_type=PluginType.EXTRACTORS,
            plugin_name="tap-mock",
            variant="meltano",
        )

        result, source = project.plugins.find_parent(tap)
        assert source == DefinitionSource.LOCKFILE
        assert result == expected
        assert result.settings == expected.settings
        assert result.settings[-1].name == "foo"

    @pytest.mark.order(2)
    def test_get_parent_no_source_enabled(
        self,
        project: Project,
        tap: ProjectPlugin,
    ) -> None:
        with project.plugins.use_preferred_source(DefinitionSource.NONE), pytest.raises(
            PluginDefinitionNotFoundError,
        ):
            project.plugins.get_parent(tap)

    def test_get_parent_no_lockfiles(
        self,
        project: Project,
        tap,
        alternative_tap,
        inherited_tap,
        alternative_target,
    ) -> None:
        # The behavior being tested here assumes that no lockfiles exist.
        shutil.rmtree(project.plugins.project.root_dir("plugins"), ignore_errors=True)
        # name="tap-mock", variant="meltano"
        # Shadows base plugin with correct variant
        parent = project.plugins.get_parent(tap)
        base = project.hub_service.find_base_plugin(
            plugin_type=PluginType.EXTRACTORS,
            plugin_name="tap-mock",
            variant="meltano",
        )
        assert base.name == parent.name
        assert base.type == parent.type

        # name="tap-mock-inherited", inherit_from="tap-mock"
        # Inherits from project plugin
        assert project.plugins.get_parent(inherited_tap) == tap

        # name="tap-mock--singer-io", inherit_from="tap-mock", variant="singer-io"
        # Inherits from base plugin with correct variant
        parent = project.plugins.get_parent(alternative_tap)
        base = project.hub_service.find_base_plugin(
            plugin_type=PluginType.EXTRACTORS,
            plugin_name="tap-mock",
            variant="singer-io",
        )
        assert base.name == parent.name
        assert base.type == parent.type

        # name="target-mock-alternative", inherit_from="target-mock"
        # Inherits from base plugin because no plugin shadowing the base plugin exists
        base = project.hub_service.find_base_plugin(
            plugin_type=PluginType.LOADERS,
            plugin_name="target-mock",
        )
        parent = project.plugins.get_parent(alternative_target)
        assert base.name == parent.name
        assert base.type == parent.type

        nonexistent_parent = ProjectPlugin(
            PluginType.EXTRACTORS,
            name="tap-foo",
            inherit_from="tap-bar",
        )
        with pytest.raises(PluginDefinitionNotFoundError) as excinfo:
            assert project.plugins.get_parent(nonexistent_parent)

        assert isinstance(excinfo.value.__cause__, PluginParentNotFoundError)

    def test_update_plugin(self, project: Project, tap) -> None:
        # update a tap with a random value
        tap.config["test"] = 42
        tap, outdated = project.plugins.update_plugin(tap)
        assert (
            project.plugins.get_plugin(tap).config["test"] == 42  # (OK magic number)
        )

        # revert back
        project.plugins.update_plugin(outdated)
        assert (
            project.plugins.get_plugin(tap).config == {}  # (OK compare with falsy)
        )

    def test_update_plugin_not_found(self, project: Project) -> None:
        nonexistent_plugin = ProjectPlugin(
            PluginType.EXTRACTORS,
            name="tap-foo",
        )

        with pytest.raises(PluginNotFoundError):
            project.plugins.update_plugin(nonexistent_plugin)

    def test_find_plugins_by_mapping_name(self, project: Project, mapper) -> None:
        assert project.plugins.find_plugins_by_mapping_name("mock-mapping-1") == [
            mapper,
        ]
        assert project.plugins.find_plugins_by_mapping_name("mock-mapping-0") == [
            mapper,
        ]
        with pytest.raises(PluginNotFoundError):
            project.plugins.find_plugins_by_mapping_name("non-existent-mapping")

    def test_find_plugins(self, project: Project, mapper) -> None:
        assert project.plugins.find_plugin("mock-mapping-1") == mapper
        assert project.plugins.find_plugin("mock-mapping-0") == mapper
        with pytest.raises(PluginNotFoundError):
            project.plugins.find_plugin("non-existent-mapping")



"""Meltano file plugin type."""

from __future__ import annotations

import typing as t

import structlog

from meltano.core.behavior.hookable import hook
from meltano.core.plugin import BasePlugin, PluginType
from meltano.core.plugin.settings_service import PluginSettingsService
from meltano.core.plugin_install_service import (
    PluginInstallReason,
    PluginInstallService,
)
from meltano.core.setting_definition import SettingDefinition, SettingKind
from meltano.core.venv_service import VirtualEnv

if t.TYPE_CHECKING:
    from os import PathLike
    from pathlib import Path

    from meltano.core.plugin.project_plugin import ProjectPlugin
    from meltano.core.project import Project


logger = structlog.getLogger(__name__)


class FilePlugin(BasePlugin):
    """Meltano file plugin type."""

    __plugin_type__ = PluginType.FILES

    EXTRA_SETTINGS: t.ClassVar[list[SettingDefinition]] = [
        SettingDefinition(
            name="_update",
            kind=SettingKind.OBJECT,
            aliases=["update"],
            value={},
        ),
    ]

    def is_invokable(self) -> bool:
        """Return whether the plugin is invokable.

        Returns:
            Whether the plugin is invokable.
        """
        return False

    def should_add_to_file(self) -> bool:
        """Return whether the plugin should be added to `meltano.yml`.

        Returns:
            True if the plugin should be added to `meltano.yml`, False otherwise.
        """
        return len(self.extras.get("update", [])) > 0

    def file_contents(self, project: Project) -> dict[Path, str]:
        """Return the contents of the files to be created or updated.

        Args:
            project: The Meltano project.

        Returns:
            A dictionary of file names and their contents.
        """
        venv = VirtualEnv(project.plugin_dir(self, "venv"))
        bundle_dir = venv.site_packages_dir / "bundle"

        return {
            path.relative_to(bundle_dir): path.read_text()
            for path in bundle_dir.glob("**/*")
            if path.is_file()
            and "__pycache__" not in path.parts
            and path != bundle_dir.joinpath("__init__.py")
        }

    def update_file_header(self, relative_path: PathLike) -> str:
        """Return the header to be added to the top of the file.

        Args:
            relative_path: The relative path of the file.

        Returns:
            The header to be added to the top of the file.
        """
        return "\n".join(
            (
                (
                    f"# This file is managed by the '{self.name}' "
                    f"{self.type.descriptor} and updated automatically when "
                    "`meltano upgrade` is run."
                ),
                (
                    f"# To prevent any manual changes from being overwritten, "
                    f"remove the {self.type.descriptor} from `meltano.yml` or "
                    "disable automatic updates:"
                ),
                (
                    f"#     meltano config --plugin-type={self.type} "
                    f"{self.name} set _update {relative_path} false"
                ),
            ),
        )

    def project_file_contents(
        self,
        project: Project,
        paths_to_update: list[str],
    ) -> dict[Path, str]:
        """Return the contents of the files to be created or updated in the project.

        Args:
            project: The Meltano project.
            paths_to_update: The paths of the files to be updated.

        Returns:
            A dictionary of file names and their contents.
        """

        def with_update_header(content: str, relative_path: PathLike):  # noqa: ANN202
            return (
                "\n\n".join([self.update_file_header(relative_path), content])
                if any(relative_path.match(path) for path in paths_to_update)
                else content
            )

        return {
            relative_path: with_update_header(content, relative_path)
            for relative_path, content in self.file_contents(project).items()
        }

    def write_file(
        self,
        project: Project,
        relative_path: PathLike,
        content: str,
    ) -> bool:
        """Write the file to the project.

        Args:
            project: The Meltano project.
            relative_path: The relative path of the file.
            content: The contents of the file.

        Returns:
            True if the file was written, False otherwise.
        """
        project_path = project.root_dir(relative_path)
        if project_path.exists() and project_path.read_text() == content:
            return False

        project_path.parent.mkdir(parents=True, exist_ok=True)
        project_path.write_text(content)

        return True

    def write_files(
        self,
        project: Project,
        files_content: dict[Path, str],
    ) -> list[Path]:
        """Write the files to the project.

        Args:
            project: The Meltano project.
            files_content: A dictionary of file names and their contents.

        Returns:
            A list of the paths of the files that were written.
        """
        return [
            relative_path
            for relative_path, content in files_content.items()
            if self.write_file(project, relative_path, content)
        ]

    def files_to_create(
        self,
        project: Project,
        paths_to_update: list[str],
    ) -> dict[Path, str]:
        """Return the contents of the files to be created in the project.

        Args:
            project: The Meltano project.
            paths_to_update: The paths of the files to be updated.

        Returns:
            A dictionary of file names and their contents.
        """

        def rename_if_exists(relative_path: Path):  # noqa: ANN202
            if not project.root_dir(relative_path).exists():
                return relative_path

            logger.info(
                f"File {str(relative_path)!r} already exists, keeping both versions",  # noqa: G004
            )
            return relative_path.with_name(
                f"{relative_path.stem} ({self.name}){relative_path.suffix}",
            )

        return {
            rename_if_exists(relative_path): content
            for relative_path, content in self.project_file_contents(
                project,
                paths_to_update,
            ).items()
        }

    def files_to_update(
        self,
        project: Project,
        paths_to_update: list[str],
    ) -> dict[Path, str]:
        """Return the contents of the files to be updated in the project.

        Args:
            project: The Meltano project.
            paths_to_update: The paths of the files to be updated.

        Returns:
            A dictionary of file names and their contents.
        """
        file_contents = self.project_file_contents(project, paths_to_update)
        return {
            relative_path: content
            for relative_path, content in file_contents.items()
            if any(relative_path.match(path) for path in paths_to_update)
        }

    def create_files(
        self,
        project: Project,
        paths_to_update: list[str] | None = None,
    ) -> list[Path]:
        """Create the files in the project.

        Args:
            project: The Meltano project.
            paths_to_update: Optional paths of the files to be updated.

        Returns:
            A list of the paths of the files that were created.
        """
        return self.write_files(
            project,
            self.files_to_create(
                project,
                [] if paths_to_update is None else paths_to_update,
            ),
        )

    def update_files(
        self,
        project: Project,
        paths_to_update: list[str] | None = None,
    ) -> list[Path]:
        """Update the files in the project.

        Args:
            project: The Meltano project.
            paths_to_update: Optional paths of the files to be updated.

        Returns:
            A list of the paths of the files that were updated.
        """
        return self.write_files(
            project,
            self.files_to_update(
                project,
                [] if paths_to_update is None else paths_to_update,
            ),
        )

    @hook("after_install")
    async def after_install(
        self,
        install_service: PluginInstallService,
        plugin: ProjectPlugin,
        reason: PluginInstallReason,
    ) -> None:
        """Trigger after install tasks.

        Args:
            install_service: The plugin installation service.
            plugin: The installed plugin.
            reason: The reason for the installation.
        """
        update_config = PluginSettingsService(install_service.project, plugin).get(
            "_update",
        )
        paths_to_update = [
            path for path, to_update in update_config.items() if to_update
        ]

        if reason is PluginInstallReason.ADD:
            logger.info(f"Adding '{plugin.name}' files to project...")  # noqa: G004
            for path in self.create_files(install_service.project, paths_to_update):
                logger.info(f"Created {path}")  # noqa: G004
        elif reason is PluginInstallReason.UPGRADE:
            logger.info(f"Updating '{plugin.name}' files in project...")  # noqa: G004
            updated_paths = self.update_files(install_service.project, paths_to_update)
            if not updated_paths:
                logger.info("Nothing to update")
                return
            for path in updated_paths:
                logger.info(f"Updated {path}")  # noqa: G004
        else:
            logger.info(
                "Run `meltano upgrade files` to update your project's "  # noqa: G004
                f"{plugin.name!r} files.",
            )



"""Fixtures to interact with Snowplow Micro."""

from __future__ import annotations

import json
import os
import subprocess
import typing as t
from urllib.request import urlopen

import backoff
import pytest

from meltano.core.project_settings_service import ProjectSettingsService


class SnowplowMicro:
    """Wrapper around the Snowplow Micro REST API."""

    def __init__(self, collector_endpoint: str):
        """Initialize the `SnowplowMicro` instance.

        Args:
            collector_endpoint: The HTTP URL to the Snowplow Micro service.
        """
        self.collector_endpoint = collector_endpoint
        self.url = f"{collector_endpoint}/micro"
        self.all()  # Wait until a connection is established

    @backoff.on_exception(backoff.expo, ConnectionError, max_tries=5)
    def get(self, endpoint: str) -> t.Any:
        with urlopen(f"{self.url}/{endpoint}") as response:
            return json.load(response)

    def all(self) -> dict[str, int]:
        """Get a dict counting the # of good/bad events, and the total # of events."""
        return self.get("all")

    def good(self) -> list[dict[str, t.Any]]:
        """Get a list of good events."""
        return self.get("good")

    def bad(self) -> list[dict[str, t.Any]]:
        """Get a list of bad events (e.g. those which failed schema validation)."""
        return self.get("bad")

    def reset(self) -> None:
        """Delete all data stored by Snowplow Micro."""
        self.get("reset")


@pytest.fixture(scope="session")
def snowplow_session(request) -> SnowplowMicro | None:
    """Start a Snowplow Micro Docker container, then yield a `SnowplowMicro` instance.

    The environment variable `$MELTANO_SNOWPLOW_COLLECTOR_ENDPOINTS` is set to
    a list containing only the collector endpoint exposed by Snowplow Micro in
    Docker.

    Yields:
        A `SnowplowMicro` instance which will collect events fired within
        tests, and can be queried to obtain info about the fired events, or
        `None` if the creation of a `SnowplowMicro` instance failed.
    """
    try:
        # Getting the `docker_services` fixture essentially causes
        # `docker-compose up` to be run
        request.getfixturevalue("docker_services")
        args = ("docker", "port", f"pytest{os.getpid()}-snowplow-1")
        proc = subprocess.run(args, capture_output=True, text=True)
        address_and_port = proc.stdout.strip().split(" -> ")[1]
        collector_endpoint = f"http://{address_and_port}"
        yield SnowplowMicro(collector_endpoint)
    except Exception:  # pragma: no cover
        yield None


@pytest.fixture()
def snowplow_optional(
    snowplow_session: SnowplowMicro | None,
    monkeypatch,
) -> SnowplowMicro | None:
    """Provide a clean `SnowplowMicro` instance.

    This fixture resets the `SnowplowMicro` instance, and enables the
    `send_anonymous_usage_stats` setting.

    Yields:
        A freshly reset `SnowplowMicro` instance, or `None` if it could not be created.
    """
    if snowplow_session is None:  # pragma: no cover
        yield None
    else:
        if isinstance(ProjectSettingsService.config_override, dict):
            monkeypatch.delitem(
                ProjectSettingsService.config_override,
                "send_anonymous_usage_stats",
                raising=False,
            )
        monkeypatch.setenv("MELTANO_SEND_ANONYMOUS_USAGE_STATS", "True")
        monkeypatch.setenv(
            "MELTANO_SNOWPLOW_COLLECTOR_ENDPOINTS",
            f'["{snowplow_session.collector_endpoint}"]',
        )
        try:
            yield snowplow_session
        finally:
            snowplow_session.reset()


@pytest.fixture()
def snowplow(snowplow_optional: SnowplowMicro | None) -> SnowplowMicro:
    """Provide a clean `SnowplowMicro` instance.

    This fixture resets the `SnowplowMicro` instance, and enables the
    `send_anonymous_usage_stats` setting.

    Returns:
        A freshly reset `SnowplowMicro` instance.
    """
    if snowplow_optional is None:  # pragma: no cover
        pytest.skip("Unable to start Snowplow Micro")
    return snowplow_optional



"""StateStoreManagers for filesystems."""

# ruff: noqa: PTH107, PTH118, PTH119, PTH120, PTH207

from __future__ import annotations

import glob
import os
import re
import shutil
import typing as t
from abc import abstractmethod, abstractproperty
from base64 import b64decode, b64encode
from contextlib import contextmanager
from datetime import datetime, timedelta, timezone
from functools import reduce
from pathlib import Path
from time import sleep
from urllib.parse import urlparse

import structlog
from smart_open import open

from meltano.core.job_state import JobState
from meltano.core.state_store.base import StateStoreManager
from meltano.core.utils import remove_suffix

if t.TYPE_CHECKING:
    from collections.abc import Iterator
    from io import TextIOWrapper

logger = structlog.stdlib.get_logger(__name__)


class InvalidStateBackendConfigurationException(Exception):
    """State backend configuration is invalid."""


class BaseFilesystemStateStoreManager(StateStoreManager):
    """Base class for filesystem state backends."""

    delimiter = "/"

    def __init__(self, uri: str, lock_timeout_seconds: int, **kwargs):  # noqa: ANN003
        """Initialize the BaseFilesystemStateStoreManager.

        Args:
            uri: the uri for the state backend
            lock_timeout_seconds: how many seconds a lock should be considered active
            kwargs: additional keyword args to pass to parent
        """
        super().__init__(**kwargs)
        self.uri = uri
        self.lock_timeout_seconds = lock_timeout_seconds
        self.parsed = urlparse(self.uri)

    def join_path(self, *components: str) -> str:
        """Join path components in filesystem-independent manner.

        Creates consistent paths for use in various cloud backends.

        Args:
            components: the path components to join

        Returns:
            components joined by '/'
        """
        return reduce(
            lambda comp1, comp2: f"{comp1}{comp2}"
            if (comp1.endswith(self.delimiter) or comp2.startswith(self.delimiter))
            else f"{comp1}{self.delimiter}{comp2}",
            components,
        )

    @staticmethod
    @abstractmethod
    def is_file_not_found_error(err: Exception) -> bool:
        """Check if err is equivalent to file not being found.

        Args:
            err: the error to check
        """
        ...

    def uri_with_path(self, path: str) -> str:
        """Build uri with the given path included.

        Args:
            path: the path to join to the uri

        Returns:
            Full URI with path included
        """
        return self.join_path(remove_suffix(self.uri, (self.state_dir)), path)

    @contextmanager
    def get_reader(self, path: str) -> Iterator[TextIOWrapper]:
        """Get reader for given path.

        Args:
            path: the path to get reader for.

        Yields:
            A TextIOWrapper to read the file/blob.
        """
        if self.client:
            with open(
                self.uri_with_path(path),
                transport_params={"client": self.client},
            ) as reader:
                yield reader
        else:
            with open(
                self.uri_with_path(path),
            ) as reader:
                yield reader

    @contextmanager
    def get_writer(self, path: str) -> Iterator[TextIOWrapper]:
        """Get writer for given path.

        Args:
            path: the path to get writer for.

        Yields:
            A TextIOWrapper to read the file/blob.
        """
        try:
            with open(
                self.uri_with_path(path),
                "w+",
                transport_params={"client": self.client} if self.client else {},
            ) as writer:
                yield writer
        except NotImplementedError:
            with open(
                self.uri_with_path(path),
                "w",
                transport_params={"client": self.client} if self.client else {},
            ) as writer:
                yield writer

    @abstractproperty
    def client(self):  # noqa: ANN201
        """Get a client for performing fs operations.

        Used for cloud backends, particularly in deleting and listing blobs.
        """
        ...

    @abstractproperty
    def state_dir(self) -> str:
        """Get the path (either filepath or prefix) that state should be stored at."""
        ...

    def get_path(self, state_id: str, filename: str | None = None) -> str:
        """Get the path for the given state_id and filename.

        Args:
            state_id: the state_id to get path for
            filename: the name of the file to get path for

        Returns:
            The constructed path.
        """
        return (
            self.join_path(self.state_dir, state_id, filename)
            if filename
            else self.join_path(self.state_dir, state_id)
        )

    def get_state_path(self, state_id: str) -> str:
        """Get the path to the file/blob storing complete state for the given state_id.

        Args:
            state_id: the state_id to get path for

        Returns:
            the path to the file/blob storing complete state for the given state_id.
        """
        return self.get_path(state_id, filename="state.json")

    def get_state_dir(self, state_id: str) -> str:
        """Get the path to the state directory for the given state_id.

        Args:
            state_id: the state_id to get path for

        Returns:
            The path to the directory for the given state_id.
        """
        return self.get_path(state_id)

    def get_lock_path(self, state_id: str) -> str:
        """Get the path to the lock file for the given state_id.

        Args:
            state_id: the state_id to get path for

        Returns:
            The path to the lock file for the given state_id.
        """
        return self.get_path(state_id, filename="lock")

    def is_locked(self, state_id: str) -> bool:
        """Indicate whether the given state_id is currently locked.

        Args:
            state_id: the state_id to check

        Returns:
            True if locked, else False

        Raises:
            Exception: if error not indicating file is not found is thrown
        """
        lock_path = self.get_lock_path(state_id)
        try:
            with self.get_reader(lock_path) as reader:
                locked_at = datetime.fromtimestamp(
                    float(reader.read()),
                    tz=timezone.utc,
                )
                if locked_at and locked_at < (
                    datetime.now(timezone.utc)
                    - timedelta(
                        seconds=self.lock_timeout_seconds,
                    )
                ):
                    self.delete(lock_path)
                    return False
                return True
        except Exception as e:
            if self.is_file_not_found_error(e):
                return False
            raise e

    def create_state_id_dir_if_not_exists(self, state_id: str) -> None:
        """Create the directory or prefix for a given state_id.

        Does nothing, but not @abstractmethod because many state backends
        will automatically create prefixes when files/blobs are created.

        Args:
            state_id: the state_id to create the dir/prefix for
        """

    @contextmanager
    def acquire_lock(self, state_id: str, retry_seconds: int = 1) -> Iterator[None]:
        """Context manager for locking state_id during reads and writes.

        Args:
            state_id: the state_id to lock.
            retry_seconds: seconds to wait between retries

        Yields:
            None
        """
        lock_path = self.get_lock_path(state_id)
        try:
            self.create_state_id_dir_if_not_exists(state_id)

            while self.is_locked(state_id):
                sleep(retry_seconds)
            with self.get_writer(lock_path) as writer:
                writer.write(str(datetime.now(timezone.utc).timestamp()))
            yield
        finally:
            self.delete(lock_path)

    @abstractmethod
    def get_state_ids(self, pattern: str | None = None) -> list[str]:
        """Get list of state_ids stored in the backend.

        Args:
            pattern: glob-style pattern to filter state_ids by

        Returns:
            List of state_ids
        """
        ...

    def get(self, state_id: str) -> JobState | None:
        """Get current state for the given state_id.

        Args:
            state_id: the state_id to get state fore.

        Returns:
            Current state, if any exists, else None

        Raises:
            Exception: if error not indicating file is not found is thrown
        """
        logger.info(f"Reading state from {self.label}")  # noqa: G004
        with self.acquire_lock(state_id):
            try:
                with self.get_reader(self.get_state_path(state_id)) as reader:
                    return JobState.from_file(state_id, reader)
            except Exception as e:
                if self.is_file_not_found_error(e):
                    logger.info(f"No state found for {state_id}.")  # noqa: G004
                    return None
                raise e

    def set(self, state: JobState) -> None:
        """Set state for the given state_id.

        Args:
            state: the state to set

        Raises:
            Exception: if error not indicating file is not found is thrown
        """
        logger.info(f"Writing state to {self.label}")  # noqa: G004
        filepath = self.get_state_path(state.state_id)
        with self.acquire_lock(state.state_id):
            if state.is_complete():
                state_to_write = state
            else:
                try:
                    with self.get_reader(filepath) as current_state_reader:
                        current_state = JobState.from_file(
                            state.state_id,
                            current_state_reader,
                        )
                        current_state.merge_partial(state)
                        state_to_write = current_state
                except Exception as e:
                    if self.is_file_not_found_error(e):
                        state_to_write = state
                    raise e
            with self.get_writer(filepath) as writer:
                writer.write(state_to_write.json())

    @abstractmethod
    def delete(self, file_or_dir_path: str):  # noqa: ANN201
        """Delete the file/blob/directory/prefix at the given path.

        Args:
            file_or_dir_path: the path to delete.
        """
        ...

    def clear(self, state_id: str) -> None:
        """Clear state for the given state_id.

        Args:
            state_id: the state_id to clear state for.
        """
        with self.acquire_lock(state_id):
            self.delete(self.get_state_path(state_id))


class LocalFilesystemStateStoreManager(BaseFilesystemStateStoreManager):
    """State backend for local filesystem."""

    label: str = "Local Filesystem"

    def __init__(self, **kwargs) -> None:  # noqa: ANN003
        """Initialize the LocalFilesystemStateStoreManager.

        Args:
            kwargs: additional kwargs to pass to parent __init__.
        """
        super().__init__(**kwargs)
        self._state_path = self.parsed.path
        Path(self._state_path).mkdir(parents=True, exist_ok=True)

    @staticmethod
    def is_file_not_found_error(err: Exception) -> bool:
        """Check if err is equivalent to file not being found.

        Args:
            err: the err to check

        Returns:
            True if error represents file not being found, else False
        """
        return isinstance(err, FileNotFoundError)

    @property
    def client(self) -> None:
        """Get a client for performing fs operations.

        Returns:
            None
        """
        return None

    @property
    def state_dir(self) -> str:
        """Get the path that state should be stored at.

        Returns:
            The relevant path
        """
        return self._state_path

    @staticmethod
    def join_path(*components: str) -> str:
        """Join path components in filesystem-dependent manner.

        Overrides base join_path method in favor of os.path.join()

        Args:
            components: the path components to join

        Returns:
            joined path
        """
        return os.path.join(*components)

    def create_state_id_dir_if_not_exists(self, state_id: str) -> None:
        """Create the directory for a given state_id.

        Args:
            state_id: the state_id to create the dir for
        """
        Path(self.get_state_dir(state_id)).mkdir(parents=True, exist_ok=True)

    def get_state_ids(self, pattern: str | None = None):  # noqa: ANN201
        """Get list of state_ids stored in the backend.

        Args:
            pattern: glob-style pattern to filter state_ids by

        Returns:
            List of state_ids
        """
        return [
            os.path.basename(os.path.dirname(state_file))
            for state_file in glob.glob(
                os.path.join(
                    self.state_dir,
                    os.path.join(pattern, "state.json")
                    if pattern
                    else os.path.join("*", "state.json"),
                ),
            )
        ]

    def delete(self, file_or_dir_path: str) -> None:
        """Delete the file/blob/directory/prefix at the given path, if it exists.

        Args:
            file_or_dir_path: the path to delete.

        Raises:
            Exception: if error not indicating file is not found is thrown
        """
        try:
            shutil.rmtree(file_or_dir_path)
        except NotADirectoryError:
            os.remove(file_or_dir_path)
        except Exception as e:
            if not self.is_file_not_found_error(e):
                raise e

    def clear(self, state_id: str) -> None:
        """Clear state for the given state_id.

        Args:
            state_id: the state_id to clear state for.
        """
        super().clear(state_id)
        shutil.rmtree(self.get_state_dir(state_id))


class WindowsFilesystemStateStoreManager(LocalFilesystemStateStoreManager):
    """State backend for local Windows filesystem."""

    label: str = "Local Windows Filesystem"
    delimiter = "\\"

    def __init__(self, **kwargs) -> None:  # noqa: ANN003
        """Initialize the LocalFilesystemStateStoreManager.

        Args:
            kwargs: additional kwargs to pass to parent __init__.
        """
        super().__init__(**kwargs)
        self._state_path = self.parsed.netloc
        Path(self._state_path).mkdir(parents=True, exist_ok=True)

    def get_path(self, state_id: str, filename: str | None = None) -> str:
        """Get the path for the given state_id and filename.

        Args:
            state_id: the state_id to get path for
            filename: the name of the file to get path for

        Returns:
            The constructed path.
        """
        state_id = b64encode(state_id.encode()).decode()
        return (
            self.join_path(self.state_dir, state_id, filename)
            if filename
            else self.join_path(self.state_dir, state_id)
        )

    def get_state_ids(self, pattern: str | None = None):  # noqa: ANN201
        """Get list of state_ids stored in the backend.

        Args:
            pattern: glob-style pattern to filter state_ids by

        Returns:
            List of state_ids
        """
        state_ids = set()
        pattern_re = re.compile(pattern.replace("*", ".*")) if pattern else None

        for state_file in glob.glob(
            os.path.join(
                self.state_dir,
                os.path.join("*", "state.json"),
            ),
        ):
            state_id = b64decode(
                os.path.basename(os.path.dirname(state_file)).encode(),
            ).decode()
            if pattern_re is None or pattern_re.match(state_id):
                state_ids.add(state_id)
        return state_ids


class CloudStateStoreManager(BaseFilesystemStateStoreManager):
    """Base class for cloud storage state store managers."""

    def __init__(self, prefix: str | None = None, **kwargs):  # noqa: ANN003
        """Initialize the CloudStateStoreManager.

        Args:
            prefix: the prefix to use for state storage
            kwargs: additional kwargs to pass to parent __init__.
        """
        super().__init__(**kwargs)
        self.prefix = prefix or self.parsed.path

    @property
    def state_dir(self) -> str:
        """Get the prefix that state should be stored at.

        Returns:
            The relevant prefix
        """
        return self.prefix.lstrip(self.delimiter).rstrip(self.delimiter)

    def uri_with_path(self, path: str) -> str:
        """Build uri with the given path included.

        Args:
            path: the path to join to the uri

        Returns:
            Full URI with path included
        """
        return self.join_path(remove_suffix(self.uri, self.prefix), path)

    @abstractmethod
    def list_all_files(self, *, with_prefix: bool = True) -> Iterator[str]:
        """List all files in the backend.

        Args:
            with_prefix: Whether to include the prefix in the lookup.

        Yields:
            The next file in the backend.
        """
        ...

    @abstractmethod
    def copy_file(self, src: str, dst: str) -> None:
        """Copy a file from one location to another.

        Args:
            src: the source path
            dst: the destination path
        """
        ...

    def get_state_ids(self, pattern: str | None = None):  # noqa: ANN201
        """Get list of state_ids stored in the backend.

        Args:
            pattern: glob-style pattern to filter state_ids by

        Returns:
            List of state_ids
        """
        if pattern:
            pattern_re = re.compile(pattern.replace("*", ".*"))
        state_ids = set()
        for filepath in self.list_all_files():
            if "/" not in filepath:
                continue

            (state_id, filename) = filepath.split("/")[-2:]
            if filename == "state.json" and (
                (not pattern) or pattern_re.match(state_id)
            ):
                state_ids.add(state_id)
        return list(state_ids)



"""Config management CLI."""

from __future__ import annotations

import json
import tempfile
import typing as t
from functools import wraps
from pathlib import Path

import click
import dotenv
import structlog

from meltano.cli.interactive import InteractiveConfig
from meltano.cli.params import InstallPlugins, get_install_options, pass_project
from meltano.cli.utils import (
    CliEnvironmentBehavior,
    CliError,
    InstrumentedGroup,
    PartialInstrumentedCmd,
)
from meltano.core.db import project_engine
from meltano.core.plugin import PluginType
from meltano.core.plugin.error import PluginNotFoundError
from meltano.core.plugin.settings_service import PluginSettingsService
from meltano.core.plugin_install_service import PluginInstallReason
from meltano.core.plugin_invoker import PluginInvoker
from meltano.core.plugin_test_service import PluginTestServiceFactory
from meltano.core.settings_service import SettingValueStore
from meltano.core.settings_store import StoreNotSupportedError
from meltano.core.tracking.contexts import CliEvent, PluginsTrackingContext
from meltano.core.utils import run_async

if t.TYPE_CHECKING:
    from meltano.core.project import Project
    from meltano.core.project_settings_service import ProjectSettingsService

logger = structlog.stdlib.get_logger(__name__)

install, no_install, only_install = get_install_options(include_only_install=True)


def _get_ctx_arg(*args: t.Any) -> click.core.Context:
    """Get the click.core.Context arg from a set of args.

    Args:
        args: the args to get Context from

    Returns:
        The click.core.Context arg.

    Raises:
        ValueError: if there is no click.core.Context in the given args.
    """
    for arg in args:
        if isinstance(arg, click.core.Context):
            return arg
    raise ValueError("No click.core.Context provided in *args")  # noqa: EM101


def _get_store_choices() -> list[SettingValueStore]:
    """Get a list of valid choices for the --store flag.

    Returns:
        SettingValueStore.writables(), without meltano_env
    """
    writables = SettingValueStore.writables()
    writables.remove(SettingValueStore.MELTANO_ENV)
    return writables


def _use_meltano_env(func):  # noqa: ANN001, ANN202
    """Override the 'meltano_yml' choice for a config command's 'store' argument.

    If an --environment flag is passed, the decorated command will use
    the MELTANO_ENV store instead of MELTANO_YML but _will not_ use MELTANO_ENV
    store if the active environment is set via the default environment.

    Args:
       func: the command to override

    Returns:
       A wrapped function with overridden store argument
    """

    @wraps(func)
    def _wrapper(*args, **kwargs):  # noqa: ANN002, ANN003, ANN202
        store = kwargs.pop("store")
        if store not in {SettingValueStore.MELTANO_YML, SettingValueStore.MELTANO_ENV}:
            return func(*args, **kwargs, store=store)
        ctx = _get_ctx_arg(*args)
        store = (
            SettingValueStore.MELTANO_YML
            if ctx.obj["is_default_environment"]
            else SettingValueStore.MELTANO_ENV
        )
        return func(*args, **kwargs, store=store)

    return _wrapper


def get_label(metadata) -> str:  # noqa: ANN001
    """Get the label for an environment variable's source.

    Args:
        metadata: the metadata for the variable

    Returns:
        string describing the source of the variable's value
    """
    source = metadata["source"]
    try:
        return f"from the {metadata['env_var']} variable in {source.label}"
    except KeyError:
        return f"from {source.label}"


@click.group(
    cls=InstrumentedGroup,
    invoke_without_command=True,
    short_help="Display Meltano or plugin configuration.",
    environment_behavior=CliEnvironmentBehavior.environment_optional_ignore_default,
)
@click.option(
    "--plugin-type",
    type=click.Choice(PluginType.cli_arguments()),
    default=None,
)
@click.argument("plugin_name")
@click.option(
    "--format",
    "config_format",
    type=click.Choice(["json", "env"]),
    default="json",
)
@click.option("--extras", is_flag=True, help="View or list only plugin extras.")
@click.option(
    "--safe/--unsafe",
    default=True,
    show_default=True,
    help="Expose values for sensitive settings.",
)
@pass_project(migrate=True)
@click.pass_context
def config(
    ctx,  # noqa: ANN001
    project: Project,
    *,
    plugin_type: str,
    plugin_name: str,
    config_format: str,
    extras: bool,
    safe: bool,
) -> None:
    """Display Meltano or plugin configuration.

    \b
    Read more at https://docs.meltano.com/reference/command-line-interface#config
    """  # noqa: D301
    tracker = ctx.obj["tracker"]
    try:
        ptype = PluginType.from_cli_argument(plugin_type) if plugin_type else None
    except ValueError:
        tracker.track_command_event(CliEvent.aborted)
        raise

    try:
        plugin = project.plugins.find_plugin(
            plugin_name,
            plugin_type=ptype,
            configurable=True,
        )
    except PluginNotFoundError:
        if plugin_name == "meltano":
            plugin = None
        else:
            tracker.track_command_event(CliEvent.aborted)
            raise

    if plugin:
        tracker.add_contexts(PluginsTrackingContext([(plugin, None)]))
    tracker.track_command_event(CliEvent.inflight)

    _, Session = project_engine(project)  # noqa: N806
    session = Session()
    try:
        if plugin:
            settings = PluginSettingsService(project, plugin)
            invoker = PluginInvoker(project, plugin)
        else:
            settings = project.settings
            invoker = None

        ctx.obj["settings"] = settings
        ctx.obj["session"] = session
        ctx.obj["invoker"] = invoker
        ctx.obj["safe"] = safe

        if ctx.invoked_subcommand is None:
            if config_format == "json":
                process = extras is not True
                json_config = settings.as_dict(
                    extras=extras,
                    process=process,
                    session=session,
                    redacted=safe,
                    redacted_value="*****",
                )
                click.echo(json.dumps(json_config, indent=2))
            elif config_format == "env":
                env = settings.as_env(
                    extras=extras,
                    session=session,
                    redacted=safe,
                    redacted_value="*****",
                )

                with tempfile.TemporaryDirectory() as temp_dir:
                    path = Path(temp_dir) / ".env"
                    for key, value in env.items():
                        dotenv.set_key(path, key, value)

                    dotenv_content = path.read_text()

                click.echo(dotenv_content)
    except Exception:
        tracker.track_command_event(CliEvent.failed)
        raise
    finally:
        session.close()


@_use_meltano_env
@config.command(
    cls=PartialInstrumentedCmd,
    name="list",
    short_help=(
        "List all settings for the specified plugin with their names, "
        "environment variables, and current values."
    ),
)
@click.option("--extras", is_flag=True)
@click.pass_context
def list_settings(ctx: click.Context, *, extras: bool) -> None:
    """List all settings for the specified plugin with their names, environment variables, and current values."""  # noqa: E501
    settings: ProjectSettingsService | PluginSettingsService = ctx.obj["settings"]
    session = ctx.obj["session"]
    tracker = ctx.obj["tracker"]
    safe: bool = ctx.obj["safe"]

    printed_custom_heading = False
    printed_extra_heading = extras

    # If `--extras` is not specified (`False`), we still want to load both
    # regular and extra settings, since we show custom extras.
    load_extras = True if extras else None

    full_config = settings.config_with_metadata(
        session=session,
        extras=load_extras,
        redacted=safe,
    )

    for name, config_metadata in full_config.items():
        value = config_metadata["value"]
        source = config_metadata["source"]
        setting_def = config_metadata["setting"]

        if extras:
            if not setting_def.is_extra:
                continue

            if setting_def.is_custom and not printed_custom_heading:
                click.echo()
                click.echo("Custom:")
                printed_custom_heading = True
        elif setting_def.is_extra:
            if not setting_def.is_custom:
                continue

            if not printed_extra_heading:
                click.echo()
                click.echo("Custom extras, plugin-specific options handled by Meltano:")
                printed_extra_heading = True
        elif setting_def.is_custom and not printed_custom_heading:
            click.echo()
            click.echo("Custom, possibly unsupported by the plugin:")
            printed_custom_heading = True

        click.secho(name, fg="blue", nl=False)

        env_keys = [var.definition for var in settings.setting_env_vars(setting_def)]
        click.echo(f" [env: {', '.join(env_keys)}]", nl=False)

        if source is not SettingValueStore.DEFAULT:
            default_value = setting_def.value
            if default_value is not None:
                click.echo(f" (default: {default_value!r})", nl=False)

        if source is SettingValueStore.DEFAULT:
            label = "default"
        elif source is SettingValueStore.INHERITED:
            label = f"inherited from '{settings.plugin.parent.name}'"  # type: ignore[union-attr]
        else:
            label = f"{get_label(config_metadata)}"

        redacted_with_value = safe and setting_def.is_redacted and value is not None

        current_value = click.style(
            value if redacted_with_value else f"{value!r}",
            fg="yellow" if redacted_with_value else "green",
        )

        click.echo(f" current value: {current_value}", nl=False)

        unexpanded_value = config_metadata.get("unexpanded_value")
        if not unexpanded_value or unexpanded_value == value:
            click.echo(f" ({label})")
        else:
            click.echo(f" ({label}: {unexpanded_value!r})")

        if setting_def.description:
            click.echo("\t", nl=False)
            if setting_def.label:
                click.echo(f"{setting_def.label}: ", nl=False)
            click.echo(f"{setting_def.description}")

    if docs_url := settings.docs_url:
        click.echo()
        click.echo(
            f"To learn more about {settings.label} and its settings, visit {docs_url}",
        )
    tracker.track_command_event(CliEvent.completed)


@config.command(cls=PartialInstrumentedCmd)
@click.option(
    "--store",
    type=click.Choice(_get_store_choices()),
    default=SettingValueStore.AUTO,
)
@click.pass_context
@_use_meltano_env
def reset(ctx, store) -> None:  # noqa: ANN001
    """Clear the configuration (back to defaults)."""
    store = SettingValueStore(store)

    settings = ctx.obj["settings"]
    session = ctx.obj["session"]
    tracker = ctx.obj["tracker"]

    try:
        metadata = settings.reset(store=store, session=session)
    except StoreNotSupportedError:  # pragma: no cover
        tracker.track_command_event(CliEvent.aborted)
        raise

    store = metadata["store"]
    click.secho(
        f"{settings.label.capitalize()} settings in {store.label} were reset",
        fg="green",
    )
    tracker.track_command_event(CliEvent.completed)


@config.command(cls=PartialInstrumentedCmd, name="set")
@click.option("--interactive", is_flag=True)
@click.option("--from-file", type=click.File("r"))
@click.argument("setting_name", nargs=-1)
@click.argument("value", required=False)
@click.option(
    "--store",
    type=click.Choice(_get_store_choices()),
    default=SettingValueStore.AUTO,
)
@click.pass_context
@_use_meltano_env
def set_(
    ctx: click.core.Context,
    *,
    setting_name: tuple[str, ...],
    value: t.Any,  # noqa: ANN401
    store: str,
    interactive: bool,
    from_file: t.TextIO,
) -> None:
    """Set the configurations' setting `<name>` to `<value>`."""
    if len(setting_name) == 1:
        setting_name = tuple(setting_name[0].split("."))

    interaction = InteractiveConfig(ctx=ctx, store=store, extras=False)

    if interactive:
        interaction.configure_all()
        ctx.exit()

    if from_file:
        setting_name += (value,)
        value = from_file.read().strip()

    interaction.set_value(setting_name=setting_name, value=value, store=store)


@config.command(cls=PartialInstrumentedCmd, name="test")
@pass_project(migrate=True)
@click.pass_context
@install
@no_install
@only_install
@run_async
async def test(
    ctx,  # noqa: ANN001
    project: Project,
    install_plugins: InstallPlugins,
) -> None:
    """Test the configuration of a plugin."""
    invoker = ctx.obj["invoker"]
    tracker = ctx.obj["tracker"]
    if not invoker:
        tracker.track_command_event(CliEvent.aborted)
        raise CliError("Testing of the Meltano project configuration is not supported")  # noqa: EM101

    session = ctx.obj["session"]
    plugin_test_service = PluginTestServiceFactory(invoker).get_test_service()

    await install_plugins(
        project,
        [plugin_test_service.plugin_invoker.plugin],
        reason=PluginInstallReason.AUTO,
    )

    stream_buffer_size = project.settings.get("elt.buffer_size")
    line_length_limit = stream_buffer_size // 2

    try:
        async with plugin_test_service.plugin_invoker.prepared(session):
            is_valid, detail = await plugin_test_service.validate(
                limit=line_length_limit,
            )
    except Exception:
        tracker.track_command_event(CliEvent.failed)
        raise

    if not is_valid:
        tracker.track_command_event(CliEvent.failed)
        raise CliError(
            "\n".join(
                (
                    "Plugin configuration is invalid",
                    detail or "Plugin did not emit any output",
                ),
            ),
        )

    click.secho("Plugin configuration is valid", fg="green")
    tracker.track_command_event(CliEvent.completed)


@config.command(cls=PartialInstrumentedCmd)
@click.argument("setting_name", nargs=-1, required=True)
@click.option(
    "--store",
    type=click.Choice(_get_store_choices()),
    default=SettingValueStore.AUTO,
)
@click.pass_context
@_use_meltano_env
def unset(ctx, setting_name, store) -> None:  # noqa: ANN001
    """Unset the configurations' setting called `<name>`."""
    store = SettingValueStore(store)

    settings = ctx.obj["settings"]
    session = ctx.obj["session"]
    tracker = ctx.obj["tracker"]

    path = list(setting_name)
    try:
        metadata = settings.unset(path, store=store, session=session)
    except StoreNotSupportedError:  # pragma: no cover
        tracker.track_command_event(CliEvent.aborted)
        raise

    name = metadata["name"]
    store = metadata["store"]
    click.secho(
        f"{settings.label.capitalize()} setting '{name}' in {store.label} was unset",
        fg="green",
    )

    current_value, source = settings.get_with_source(name, session=session)
    if source is not SettingValueStore.DEFAULT:
        click.secho(
            f"Current value is now: {current_value!r} ({get_label(metadata)})",
            fg="yellow",
        )
    tracker.track_command_event(CliEvent.completed)



"""Lock command."""

from __future__ import annotations

import typing as t

import click
import structlog

from meltano.cli.params import pass_project
from meltano.cli.utils import CliError, PartialInstrumentedCmd
from meltano.core.plugin import PluginType
from meltano.core.plugin_lock_service import (
    LockfileAlreadyExistsError,
    PluginLockService,
)
from meltano.core.project_plugins_service import DefinitionSource
from meltano.core.tracking.contexts import CliEvent, PluginsTrackingContext

if t.TYPE_CHECKING:
    from meltano.core.project import Project
    from meltano.core.tracking import Tracker


__all__ = ["lock"]
logger = structlog.get_logger(__name__)


@click.command(cls=PartialInstrumentedCmd, short_help="Lock plugin definitions.")
@click.option(
    "--all",
    "all_plugins",
    is_flag=True,
    help="Lock all the plugins of the project.",
)
@click.option(
    "--plugin-type",
    type=click.Choice(PluginType.cli_arguments()),
    help="Lock only the plugins of the given type.",
)
@click.argument("plugin_name", nargs=-1, required=False)
@click.option("--update", "-u", is_flag=True, help="Update the lock file.")
@click.pass_context
@pass_project()
def lock(
    project: Project,
    ctx: click.Context,
    *,
    all_plugins: bool,
    plugin_type: str | None,
    plugin_name: tuple[str, ...],
    update: bool,
) -> None:
    """Lock plugin definitions.

    \b
    Read more at https://docs.meltano.com/reference/command-line-interface#lock
    """  # noqa: D301
    tracker: Tracker = ctx.obj["tracker"]

    lock_service = PluginLockService(project)
    if (all_plugins and plugin_name) or not (all_plugins or plugin_name):
        tracker.track_command_event(CliEvent.aborted)
        raise CliError("Exactly one of --all or plugin name must be specified.")  # noqa: EM101

    try:
        with project.plugins.use_preferred_source(DefinitionSource.ANY):
            # Make it a list so source preference is not lazily evaluated.
            plugins = list(project.plugins.plugins())
    except Exception:
        tracker.track_command_event(CliEvent.aborted)
        raise

    if plugin_name:
        plugins = [plugin for plugin in plugins if plugin.name in plugin_name]

    if plugin_type:
        plugin_type = PluginType.from_cli_argument(plugin_type)
        plugins = [plugin for plugin in plugins if plugin.type == plugin_type]

    tracked_plugins = []

    if not plugins:
        tracker.track_command_event(CliEvent.aborted)
        errmsg = "No matching plugin(s) found"
        raise CliError(errmsg)

    click.echo(f"Locking {len(plugins)} plugin(s)...")
    for plugin in plugins:
        descriptor = f"{plugin.type.descriptor} {plugin.name}"
        if plugin.is_custom():
            click.secho(f"{descriptor.capitalize()} is a custom plugin", fg="yellow")
        elif plugin.inherit_from is not None:
            click.secho(
                f"{descriptor.capitalize()} is an inherited plugin",
                fg="yellow",
            )
        else:
            plugin.parent = None
            with project.plugins.use_preferred_source(DefinitionSource.HUB):
                plugin = project.plugins.ensure_parent(plugin)
            try:
                lock_service.save(plugin, exists_ok=update)
            except LockfileAlreadyExistsError as err:
                relative_path = err.path.relative_to(project.root)
                click.secho(
                    f"Lockfile exists for {descriptor} at {relative_path}",
                    fg="red",
                )
                continue

            tracked_plugins.append((plugin, None))
            click.secho(f"Locked definition for {descriptor}", fg="green")

    tracker.add_contexts(PluginsTrackingContext(tracked_plugins))
    tracker.track_command_event(CliEvent.completed)



"""Utilities to extract information from jsonschema files."""

from __future__ import annotations

import typing as t
from contextlib import suppress

if t.TYPE_CHECKING:
    from collections.abc import Generator


def meltano_config_env_locations(
    manifest_schema: dict[str, t.Any],
    env_definition_ref: str = "#/$defs/env",
) -> set[str]:
    """Find all locations within a Meltano manifest file that can have an env field.

    Args:
        manifest_schema: The jsonschema for Meltano manifests.
        env_definition_ref: The jsonschema ref that defines what is being searched for.

    Returns:
        The locations within a Meltano manifest file that can have an env
        field, as jq filter strings.
    """
    return set(JsonschemaRefLocationParser(manifest_schema, env_definition_ref).parse())


class JsonschemaRefLocationParser:
    """Minimal jsonschema parser that finds the locations of a specified ref."""

    def __init__(self, schema: dict[str, t.Any], target_ref: str) -> None:
        """Initialize the `JsonschemaRefLocationParser`.

        Args:
            schema: The schema being parsed.
            target_ref: The ref being searched for within the given schema.
        """
        self.root_schema = schema
        self.target_ref = target_ref
        self._resolved_refs: dict[str, t.Any] = {}

    def parse(
        self,
        schema: dict[str, t.Any] | None = None,
        path: tuple[str, ...] = (),
    ) -> Generator[str, None, None]:
        """Parse the jsonschema to find the locations of the target ref.

        Args:
            schema: The jsonschema to parse. Defaults to the root schema.
            path: The path taken to arrive at the provided schema. Each element
                is a jq filter that can be applied to each previous element up
                to the root of the root schema.

        Yields:
            The path argument concatenated, which is a location of the target
            ref specified as jq filters.
        """
        if schema is None:
            schema = self.root_schema

        if "$ref" in schema:
            if schema["$ref"] == self.target_ref:
                yield "".join(path)
            else:
                yield from self.parse(self.resolve_ref(schema["$ref"], path), path)
        elif "type" in schema:
            if schema["type"] == "object":
                if "properties" in schema:
                    for name, prop in schema["properties"].items():
                        yield from self.parse(prop, (*path, f".{name}"))
                elif "allOf" in schema:
                    for shard in schema["allOf"]:
                        yield from self.parse(shard, path)
            elif schema["type"] == "array" and "items" in schema:
                yield from self.parse(schema["items"], (*path, "[]"))

    def resolve_ref(self, ref: str, path: tuple[str, ...]) -> dict[str, t.Any]:
        """Resolve a local jsonschema ref into a schema.

        Args:
            ref: The jsonschema reference to resolve.
            path: The path taken to arrive at the ref being resolved. Each
                element is a jq filter that can be applied to each previous
                element up to the root of the root schema.

        Raises:
            ValueError: Attempted to resolve a non-local reference.
            RecursionError: Attempted to resolve a recursive reference.
            KeyError: A component of the path in the ref was not found.

        Returns:
            The jsonschema the provided ref refers to.
        """
        with suppress(KeyError):
            return self._resolved_refs[ref]

        location_as_ref = self.path_to_ref(path)
        if location_as_ref.startswith(ref):
            raise RecursionError(
                "Cannot resolve recursive jsonschema "  # noqa: EM102
                f"ref {ref!r} at {location_as_ref!r}",
            )

        current_schema = self.root_schema
        parts = ref.split("/")

        if parts[0] != "#":
            raise ValueError("Cannot resolve non-local jsonschema ref")  # noqa: EM101

        for key in parts[1:]:
            try:
                current_schema = current_schema[key]
            except KeyError as ex:  # noqa: PERF203
                raise KeyError(
                    f"Unable to resolve local jsonschema ref {ref!r} at level {key!r}",  # noqa: EM102
                ) from ex
        return current_schema

    @staticmethod
    def path_to_ref(path: tuple[str, ...]) -> str:
        """Convert a sequence of jq filters into a jsonschema ref.

        Args:
            path: The sequence of jq filters that specify the location within the json.

        Returns:
            A jsonschema ref to the same location as the provided path.
        """
        return "#" + "".join(path).replace(".", "/").replace("[]", "")
